"""Set of metaclasses used throughout the library."""
import datetime
from abc import ABC, abstractmethod
from collections import UserDict
from typing import Optional, List
try:
    import orjson as json
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        try:
            import simplejson as json
        except ImportError:
            import json

import esprima
from esprima.error_handler import Error as ParseError
from validator_collection import validators, checkers, errors as validator_errors

from highcharts_core import constants, errors, utility_functions
from highcharts_core.decorators import validate_types
from highcharts_core.js_literal_functions import serialize_to_js_literal, assemble_js_literal,\
    get_key_value_pairs


class HighchartsMeta(ABC):
    """Metaclass that is used to define the standard interface exposed for serializable
    objects."""

    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs.get(key, None))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        self_js_literal = self.to_js_literal()
        other_js_literal = other.to_js_literal()

        return self_js_literal == other_js_literal

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return None

    def _process_required_modules(self, scripts = None, include_extension = False) -> List[str]:
        """Return the list of URLs from which the Highcharts JavaScript modules
        needed to render the chart can be retrieved.
        
        :param scripts: Initial set of scripts that are context-dependent.
        :type scripts: :class:`list <python:list>` of :class:`str <python:str>`
        
        :param include_extension: if ``True``, will return script names with the 
          ``'.js'`` extension included. Defaults to ``False``.
        :type include_extension: :class:`bool <python:bool>`

        :rtype: :class:`list <python:list>`
        """
        if not scripts:
            scripts = []
            
        properties = [x[1:] for x in self.__dict__
                      if x.startswith('_') and hasattr(self, x[1:])]

        for property_name in properties:
            property_value = getattr(self, property_name, None)
            if property_value is None:
                continue
            if checkers.is_iterable(property_value, forbid_literals = (str, bytes, dict)):
                additional_scripts = []
                for item in property_value:
                    if hasattr(item, 'get_required_modules'):
                        item_scripts = [x for x in item.get_required_modules()
                                        if x not in scripts]
                        additional_scripts.extend(item_scripts)
                if additional_scripts:
                    scripts.extend(additional_scripts)
                    continue
            if isinstance(property_value, HighchartsMeta):
                additional_scripts = [x for x in property_value.get_required_modules()
                                      if x not in scripts]
                if additional_scripts:
                    scripts.extend(additional_scripts)
                    continue
            property_name_as_camelCase = utility_functions.to_camelCase(property_name)
            dot_path = f'{self._dot_path}.' or ''
            dot_path += property_name_as_camelCase
            scripts.extend([x for x in constants.MODULE_REQUIREMENTS.get(dot_path, [])
                            if x not in scripts])

        if include_extension:
            final_scripts = []
            for script in scripts:
                if script.endswith('.css') or script.endswith('.js'):
                    final_scripts.append(script)
                elif script.startswith('css/'):
                    final_scripts.append(f'{script}.css')
                else:
                    final_scripts.append(f'{script}.js')
            return final_scripts
            
        return scripts

    def get_required_modules(self, include_extension = False) -> List[str]:
        """Return the list of URLs from which the Highcharts JavaScript modules
        needed to render the chart can be retrieved.
        
        :param include_extension: if ``True``, will return script names with the 
          ``'.js'`` extension included. Defaults to ``False``.
        :type include_extension: :class:`bool <python:bool>`

        :rtype: :class:`list <python:list>`
        """
        initial_scripts = constants.MODULE_REQUIREMENTS.get(self._dot_path, [])
        prelim_scripts = self._process_required_modules(initial_scripts, include_extension = include_extension)
        scripts = []
        
        has_all_indicators = False
        for script in prelim_scripts:
            if script.endswith('indicators-all.js'):
                has_all_indicators = True
        
        for script in prelim_scripts:
            if '/indicators/' in script and has_all_indicators:
                continue
            scripts.append(script)
            
        return scripts

    def _untrimmed_mro_ancestors(self, in_cls = None) -> dict:
        """Walk through the parent classes and consolidate the results of their
        :meth:`_to_untrimmed_dict() <HighchartsMeta._to_untrimmed_dict__>` methods into
        a single :class:`dict <python:dict>`.

        :rtype: :class:`dict <python:dict>`
        """
        return utility_functions.mro__to_untrimmed_dict(self, in_cls = in_cls)

    @abstractmethod
    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        """Generate the first-level of the :class:`dict <python:dict>` representation of
        the object.

        .. note::

          This method does *not* traverse the object structure to convert the object into
          a full :class:`dict <python:dict>` representation - it merely goes "part" of the
          way there to replace the Highcharts for Python keys with their correpsond
          Highcharts JS keys.

        :rtype: :class:`dict <python:dict>`
        """
        raise NotImplementedError()

    @staticmethod
    def trim_iterable(untrimmed,
                      to_json = False):
        """Convert any :class:`EnforcedNullType` values in ``untrimmed`` to ``'null'``.

        :param untrimmed: The iterable whose members may still be
          :obj:`None <python:None>` or Python objects.
        :type untrimmed: iterable

        :param to_json: If ``True``, will remove all members from ``untrimmed`` that are
          not serializable to JSON. Defaults to ``False``.
        :type to_json: :class:`bool <python:bool>`

        :rtype: iterable
        """
        if not checkers.is_iterable(untrimmed, forbid_literals = (str, bytes, dict)):
            return untrimmed

        trimmed = []
        for item in untrimmed:
            if checkers.is_type(item, 'CallbackFunction') and to_json:
                continue
            elif item is None or item == constants.EnforcedNull:
                trimmed.append('null')
            elif hasattr(item, 'trim_dict'):
                untrimmed_item = item._to_untrimmed_dict()
                item_as_dict = HighchartsMeta.trim_dict(untrimmed_item, to_json = to_json)
                if item_as_dict:
                    trimmed.append(item_as_dict)
            elif isinstance(item, dict):
                if item:
                    trimmed.append(HighchartsMeta.trim_dict(item, to_json = to_json))
            elif checkers.is_iterable(item, forbid_literals = (str, bytes, dict)):
                if item:
                    trimmed.append(HighchartsMeta.trim_iterable(item, to_json = to_json))
            else:
                trimmed.append(item)

        return trimmed

    @staticmethod
    def trim_dict(untrimmed: dict,
                  to_json: bool = False) -> dict:
        """Remove keys from ``untrimmed`` whose values are :obj:`None <python:None>` and
        convert values that have ``.to_dict()`` methods.

        :param untrimmed: The :class:`dict <python:dict>` whose values may still be
          :obj:`None <python:None>` or Python objects.
        :type untrimmed: :class:`dict <python:dict>`

        :param to_json: If ``True``, will remove all keys from ``untrimmed`` that are not
          serializable to JSON. Defaults to ``False``.
        :type to_json: :class:`bool <python:bool>`

        :returns: Trimmed :class:`dict <python:dict>`
        :rtype: :class:`dict <python:dict>`
        """
        as_dict = {}
        for key in untrimmed:
            value = untrimmed.get(key, None)
            # bool -> Boolean
            if isinstance(value, bool):
                as_dict[key] = value
            # Callback Function
            elif checkers.is_type(value, 'CallbackFunction') and to_json:
                continue
            # HighchartsMeta -> dict --> object
            elif value and hasattr(value, '_to_untrimmed_dict'):
                untrimmed_value = value._to_untrimmed_dict()
                trimmed_value = HighchartsMeta.trim_dict(untrimmed_value,
                                                         to_json = to_json)
                if trimmed_value:
                    as_dict[key] = trimmed_value
            # Enforced null
            elif isinstance(value, constants.EnforcedNullType):
                as_dict[key] = 'null'
            # dict -> object
            elif isinstance(value, dict):
                trimmed_value = HighchartsMeta.trim_dict(value,
                                                         to_json = to_json)
                if trimmed_value:
                    as_dict[key] = trimmed_value
            # iterable -> array
            elif checkers.is_iterable(value, forbid_literals = (str, bytes, dict)):
                trimmed_value = HighchartsMeta.trim_iterable(value, to_json = to_json)
                if trimmed_value:
                    as_dict[key] = trimmed_value
            # Pandas Timestamp
            elif checkers.is_type(value, 'Timestamp'):
                as_dict[key] = value.timestamp()
            # other truthy -> str / number
            elif value:
                trimmed_value = HighchartsMeta.trim_iterable(value, to_json = to_json)
                if trimmed_value:
                    as_dict[key] = trimmed_value
            # other falsy -> str / number
            elif value in [0, 0., False]:
                as_dict[key] = value

        return as_dict

    @classmethod
    @abstractmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        raise NotImplementedError()

    @classmethod
    def from_dict(cls,
                  as_dict: dict,
                  allow_snake_case: bool = True):
        """Construct an instance of the class from a :class:`dict <python:dict>` object.

        :param as_dict: A :class:`dict <python:dict>` representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :param allow_snake_case: If ``True``, interprets ``snake_case`` keys as equivalent
          to ``camelCase`` keys. Defaults to ``True``.
        :type allow_snake_case: :class:`bool <python:bool>`

        :returns: A Python object representation of ``as_dict``.
        :rtype: :class:`HighchartsMeta`
        """
        as_dict = validators.dict(as_dict, allow_empty = True) or {}
        clean_as_dict = {}
        if allow_snake_case:
            clean_as_dict = {utility_functions.to_camelCase(key): as_dict[key]
                             for key in as_dict}
        else:
            clean_as_dict = {key: as_dict[key]
                             for key in as_dict}

        kwargs = cls._get_kwargs_from_dict(clean_as_dict)

        return cls(**kwargs)

    @classmethod
    def from_json(cls,
                  as_json_or_file,
                  allow_snake_case: bool = True):
        """Construct an instance of the class from a JSON string.

        :param as_json_or_file: The JSON string for the object or the filename of a file
          that contains the JSON string.
        :type as_json: :class:`str <python:str>` or :class:`bytes <python:bytes>`

        :param allow_snake_case: If ``True``, interprets ``snake_case`` keys as equivalent
          to ``camelCase`` keys. Defaults to ``True``.
        :type allow_snake_case: :class:`bool <python:bool>`

        :returns: A Python objcet representation of ``as_json``.
        :rtype: :class:`HighchartsMeta`
        """
        is_file = checkers.is_file(as_json_or_file)
        if is_file:
            with open(as_json_or_file, 'r') as file_:
                as_str = file_.read()
        else:
            as_str = as_json_or_file

        as_dict = json.loads(as_str)

        if checkers.is_iterable(as_dict, forbid_literals = (str, bytes, dict, UserDict)):
            return [cls.from_dict(x, allow_snake_case = allow_snake_case)
                    for x in as_dict]

        return cls.from_dict(as_dict,
                             allow_snake_case = allow_snake_case)

    def to_dict(self) -> dict:
        """Generate a :class:`dict <python:dict>` representation of the object compatible
        with the Highcharts JavaScript library.

        .. note::

          The :class:`dict <python:dict>` representation has a property structure and
          naming convention that is *intentionally* consistent with the Highcharts
          JavaScript library. This is not Pythonic, but it makes managing the interplay
          between the two languages much, much simpler.

        :returns: A :class:`dict <python:dict>` representation of the object.
        :rtype: :class:`dict <python:dict>`
        """
        untrimmed = self._to_untrimmed_dict()

        return self.trim_dict(untrimmed)

    def to_json(self,
                filename = None,
                encoding = 'utf-8'):
        """Generate a JSON string/byte string representation of the object compatible with
        the Highcharts JavaScript library.

        .. note::

          This method will either return a standard :class:`str <python:str>` or a
          :class:`bytes <python:bytes>` object depending on the JSON serialization library
          you are using. For example, if your environment has
          `orjson <https://github.com/ijl/orjson>`_, the result will be a
          :class:`bytes <python:bytes>` representation of the string.

        :param filename: The name of a file to which the JSON string should be persisted.
          Defaults to :obj:`None <python:None>`
        :type filename: Path-like

        :param encoding: The character encoding to apply to the resulting object. Defaults
          to ``'utf-8'``.
        :type encoding: :class:`str <python:str>`

        :returns: A JSON representation of the object compatible with the Highcharts
          library.
        :rtype: :class:`str <python:str>` or :class:`bytes <python:bytes>`
        """
        if filename:
            filename = validators.path(filename)

        untrimmed = self._to_untrimmed_dict()

        as_dict = self.trim_dict(untrimmed, to_json = True)

        for key in as_dict:
            if as_dict[key] == constants.EnforcedNull or as_dict[key] == 'null':
                as_dict[key] = None
        try:
            as_json = json.dumps(as_dict, encoding = encoding)
        except TypeError:
            as_json = json.dumps(as_dict)

        if filename:
            if isinstance(as_json, bytes):
                write_type = 'wb'
            else:
                write_type = 'w'

            with open(filename, write_type, encoding = encoding) as file_:
                file_.write(as_json)

        return as_json

    def to_js_literal(self,
                      filename = None,
                      encoding = 'utf-8') -> Optional[str]:
        """Return the object represented as a :class:`str <python:str>` containing the
        JavaScript object literal.

        :param filename: The name of a file to which the JavaScript object literal should
          be persisted. Defaults to :obj:`None <python:None>`
        :type filename: Path-like

        :param encoding: The character encoding to apply to the resulting object. Defaults
          to ``'utf-8'``.
        :type encoding: :class:`str <python:str>`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        if filename:
            filename = validators.path(filename)

        untrimmed = self._to_untrimmed_dict()
        as_dict = {}
        for key in untrimmed:
            item = untrimmed[key]
            serialized = serialize_to_js_literal(item, encoding = encoding)
            if serialized is not None:
                as_dict[key] = serialized

        as_str = assemble_js_literal(as_dict)

        if filename:
            with open(filename, 'w', encoding = encoding) as file_:
                file_.write(as_str)

        return as_str

    @classmethod
    def _validate_js_literal(cls,
                             as_str,
                             range = True,
                             _break_loop_on_failure = False):
        """Parse ``as_str`` as a valid JavaScript literal object.

        :param as_str: The string to parse as a JavaScript literal object.
        :type as_str: :class:`str <python:str>`

        :param range: If ``True``, includes location and range data for each node in the
          AST returned. Defaults to ``False``.
        :type range: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, will not loop if the method fails to
          parse/validate ``as_str``. Defaults to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: The parsed AST representation of ``as_str`` and the updated string.
        :rtype: 2-member :class:`tuple <python:tuple>` of :class:`esprima.nodes.Script`
          and :class:`str <python:str>`
        """
        if """document.addEventListener('DOMContentLoaded', function() {\n""" in as_str:
            as_str = as_str.replace(
                """document.addEventListener('DOMContentLoaded', function() {\n""", ''
            )
            as_str = as_str[:-3]
        try:
            parsed = esprima.parseScript(as_str, loc = range, range = range)
        except ParseError:
            try:
                parsed = esprima.parseModule(as_str, loc = range, range = range)
            except ParseError:
                if not _break_loop_on_failure:
                    as_str = f"""var randomVariable = {as_str}"""
                    return cls._validate_js_literal(as_str,
                                                    range = range,
                                                    _break_loop_on_failure = True)
                else:
                    raise errors.HighchartsParseError('._validate_js_literal() expects '
                                                      'a str containing a valid '
                                                      'JavaScript literal object. Could '
                                                      'not find a valid literal.')

        return parsed, as_str

    @classmethod
    def from_js_literal(cls,
                        as_str_or_file,
                        allow_snake_case: bool = True,
                        _break_loop_on_failure: bool = False):
        """Return a Python object representation of a Highcharts JavaScript object
        literal.

        :param as_str_or_file: The JavaScript object literal, represented either as a
          :class:`str <python:str>` or as a filename which contains the JS object literal.
        :type as_str_or_file: :class:`str <python:str>`

        :param allow_snake_case: If ``True``, interprets ``snake_case`` keys as equivalent
          to ``camelCase`` keys. Defaults to ``True``.
        :type allow_snake_case: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, will break any looping operations in
          the event of a failure. Otherwise, will attempt to repair the failure. Defaults
          to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: A Python object representation of the Highcharts JavaScript object
          literal.
        :rtype: :class:`HighchartsMeta`
        """
        is_file = checkers.is_file(as_str_or_file)
        if is_file:
            with open(as_str_or_file, 'r') as file_:
                as_str = file_.read()
        else:
            as_str = as_str_or_file

        parsed, updated_str = cls._validate_js_literal(as_str)

        as_dict = {}
        if not parsed.body:
            return cls()

        if len(parsed.body) > 1:
            raise errors.HighchartsCollectionError(f'each JavaScript object literal is '
                                                   f'expected to contain one object. '
                                                   f'However, you attempted to parse '
                                                   f'{len(parsed.body)} objects.')

        body = parsed.body[0]
        if not checkers.is_type(body, 'VariableDeclaration') and \
           _break_loop_on_failure is False:
            prefixed_str = f'var randomVariable = {as_str}'
            return cls.from_js_literal(prefixed_str,
                                       _break_loop_on_failure = True)
        elif not checkers.is_type(body, 'VariableDeclaration'):
            raise errors.HighchartsVariableDeclarationError('To parse a JavaScriot '
                                                            'object literal, it is '
                                                            'expected to be either a '
                                                            'variable declaration or a'
                                                            'standalone block statement.'
                                                            'Input received did not '
                                                            'conform.')
        declarations = body.declarations
        if not declarations:
            return cls()

        if len(declarations) > 1:
            raise errors.HighchartsCollectionError(f'each JavaScript object literal is '
                                                   f'expected to contain one object. '
                                                   f'However, you attempted to parse '
                                                   f'{len(parsed.body)} objects.')
        object_expression = declarations[0].init
        if not checkers.is_type(object_expression, 'ObjectExpression'):
            raise errors.HighchartsParseError(f'Highcharts expects an object literal to '
                                              f'to be defined as a standard '
                                              f'ObjectExpression. Received: '
                                              f'{type(object_expression)}')

        properties = object_expression.properties
        if not properties:
            return cls()

        key_value_pairs = [(x[0], x[1]) for x in get_key_value_pairs(properties,
                                                                     updated_str)]

        for pair in key_value_pairs:
            as_dict[pair[0]] = pair[1]

        return cls.from_dict(as_dict,
                             allow_snake_case = allow_snake_case)

    @classmethod
    def _copy_dict_key(cls,
                       key,
                       original,
                       other,
                       overwrite = True,
                       **kwargs):
        """Copies the value of ``key`` from ``original`` to ``other``.

        :param key: The key that is to be copied.
        :type key: :class:`str <python:str>`

        :param original: The original :class:`dict <python:dict>` from which it should
          be copied.
        :type original: :class:`dict <python:dict>`

        :param other: The :class:`dict <python:dict>` to which it should be copied.
        :type other: :class:`dict <python:dict>`

        :returns: The value that should be placed in ``other`` for ``key``.
        """
        original_value = original[key]
        other_value = other.get(key, None)

        if isinstance(original_value, (dict, UserDict)):
            new_value = {}
            for subkey in original_value:
                new_key_value = cls._copy_dict_key(subkey,
                                                   original_value,
                                                   other_value,
                                                   overwrite = overwrite,
                                                   **kwargs)
                new_value[subkey] = new_key_value

            return new_value

        elif checkers.is_iterable(original_value,
                                  forbid_literals = (str,
                                                     bytes,
                                                     dict,
                                                     UserDict)):
            if overwrite:
                new_value = [x for x in original_value]

                return new_value

            return other_value

        elif other_value and not overwrite:
            return other_value

        return original_value

    def copy(self,
             other = None,
             overwrite = True,
             **kwargs):
        """Copy the configuration settings from this instance to the ``other`` instance.

        :param other: The target instance to which the properties of this instance should
          be copied. If :obj:`None <python:None>`, will create a new instance and populate
          it with properties copied from ``self``. Defaults to :obj:`None <python:None>`.
        :type other: :class:`HighchartsMeta`

        :param overwrite: if ``True``, properties in ``other`` that are already set will
          be overwritten by their counterparts in ``self``. Defaults to ``True``.
        :type overwrite: :class:`bool <python:bool>`

        :param kwargs: Additional keyword arguments. Some special descendents of
          :class:`HighchartsMeta` may have special implementations of this method which
          rely on additional keyword arguments.

        :returns: A mutated version of ``other`` with new property values

        """
        if not other:
            other = self.__class__()

        if not isinstance(other, self.__class__):
            raise errors.HighchartsValueError(f'other is expected to be a '
                                              f'{self.__class__.__name__} instance. Was: '
                                              f'{other.__class__.__name__}')

        self_as_dict = self.to_dict()
        other_as_dict = other.to_dict()

        new_dict = {}
        for key in self_as_dict:
            new_dict[key] = self._copy_dict_key(key,
                                                original = self_as_dict,
                                                other = other_as_dict,
                                                overwrite = overwrite,
                                                **kwargs)

        cls = other.__class__

        other = cls.from_dict(new_dict)

        return other


class JavaScriptDict(UserDict):
    """Special :class:`dict <python:dict>` class which constructs a JavaScript
    object that can be represented as a string.

    Keys are validated to be valid variable names, while values are validated to be
    strings.

    When serialized to :class:`str <python:str>`, keys are **not** wrapped in double
    quotes (as they would be in JSON) to ensure that the resulting string can be evaluated
    as JavaScript code.

    """
    _valid_value_types = None
    _allow_empty_value = True

    def __setitem__(self, key, item):
        validate_key = False
        try:
            validate_key = key not in self
        except AttributeError:
            validate_key = True

        if validate_key:
            try:
                key = validators.variable_name(key, allow_empty = False)
            except validator_errors.InvalidVariableNameError as error:
                if '-' in key:
                    try:
                        test_key = key.replace('-', '_')
                        validators.variable_name(test_key, allow_empty = False)
                    except validator_errors.InvalidVariableNameError:
                        raise error
                else:
                    raise error

        if self._valid_value_types:
            try:
                item = validate_types(item,
                                      types = self._valid_value_types,
                                      allow_none = self._allow_empty_value)
            except errors.HighchartsValueError as error:
                if self._allow_empty_value and not item:
                    item = None
                else:
                    try:
                        item = self._valid_value_types(item)
                    except (TypeError, ValueError, AttributeError):
                        raise error

        super().__setitem__(key, item)

    @classmethod
    def from_dict(cls, as_dict):
        """Construct an instance of the class from a :class:`dict <python:dict>` object.

        :param as_dict: A :class:`dict <python:dict>` representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: A Python object representation of ``as_dict``.
        :rtype: :class:`JavaScriptDict`
        """
        as_dict = validators.dict(as_dict, allow_empty = True)
        if not as_dict:
            return cls()

        as_obj = cls()
        for key in as_dict:
            as_obj[key] = as_dict.get(key, None)

        return as_obj

    @classmethod
    def from_json(cls, as_json):
        """Construct an instance of the class from a JSON string.

        :param as_json: The JSON string for the object.
        :type as_json: :class:`str <python:str>` or :class:`bytes <python:bytes>`

        :returns: A Python objcet representation of ``as_json``.
        :rtype: :class:`HighchartsMeta`
        """
        as_dict = json.loads(as_json)

        return cls.from_dict(as_dict)

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return self.data

    def to_dict(self):
        """Generate a :class:`dict <python:dict>` representation of the object compatible
        with the Highcharts JavaScript library.

        .. note::

          The :class:`dict <python:dict>` representation has a property structure and
          naming convention that is *intentionally* consistent with the Highcharts
          JavaScript library. This is not Pythonic, but it makes managing the interplay
          between the two languages much, much simpler.

        :returns: A :class:`dict <python:dict>` representation of the object.
        :rtype: :class:`dict <python:dict>`
        """
        return self.data

    def to_json(self,
                filename = None,
                encoding = 'utf-8'):
        """Generate a JSON string/byte string representation of the object compatible with
        the Highcharts JavaScript library.

        .. note::

          This method will either return a standard :class:`str <python:str>` or a
          :class:`bytes <python:bytes>` object depending on the JSON serialization library
          you are using. For example, if your environment has
          `orjson <https://github.com/ijl/orjson>`_, the result will be a
          :class:`bytes <python:bytes>` representation of the string.

        :param filename: The name of a file to which the JSON string should be persisted.
          Defaults to :obj:`None <python:None>`
        :type filename: Path-like

        :param encoding: The character encoding to apply to the resulting object. Defaults
          to ``'utf8'``.
        :type encoding: :class:`str <python:str>`

        :returns: A JSON representation of the object compatible with the Highcharts
          library.
        :rtype: :class:`str <python:str>` or :class:`bytes <python:bytes>`
        """
        if filename:
            filename = validators.path(filename)

        as_dict = self.to_dict()
        try:
            as_json = json.dumps(as_dict, encoding = encoding)
        except TypeError:
            as_json = json.dumps(as_dict)

        if filename:
            if isinstance(as_json, bytes):
                write_type = 'wb'
            else:
                write_type = 'w'

            with open(filename, write_type, encoding = encoding) as file_:
                file_.write(as_json)

        return as_json

    def to_js_literal(self,
                      filename = None,
                      encoding = 'utf-8') -> Optional[str]:
        """Return the object represented as a :class:`str <python:str>` containing the
        JavaScript object literal.

        :param filename: The name of a file to which the JavaScript object literal should
          be persisted. Defaults to :obj:`None <python:None>`
        :type filename: Path-like

        :param encoding: The character encoding to apply to the resulting object. Defaults
          to ``'utf-8'``.
        :type encoding: :class:`str <python:str>`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        if filename:
            filename = validators.path(filename)

        untrimmed = self._to_untrimmed_dict()
        as_dict = {}
        for key in untrimmed:
            item = untrimmed[key]
            serialized = serialize_to_js_literal(item, encoding = encoding)
            if serialized is not None:
                as_dict[key] = serialized

        as_str = assemble_js_literal(as_dict, keys_as_strings = True)

        if filename:
            with open(filename, 'w', encoding = encoding) as file_:
                file_.write(as_str)

        return as_str

    @classmethod
    def _validate_js_literal(cls,
                             as_str,
                             range = True,
                             _break_loop_on_failure = False):
        """Parse ``as_str`` as a valid JavaScript literal object.

        :param as_str: The string to parse as a JavaScript literal object.
        :type as_str: :class:`str <python:str>`

        :param range: If ``True``, includes location and range data for each node in the
          AST returned. Defaults to ``False``.
        :type range: :class:`bool <python:bool>`

        :param _break_loop_on_failure: If ``True``, will not loop if the method fails to
          parse/validate ``as_str``. Defaults to ``False``.
        :type _break_loop_on_failure: :class:`bool <python:bool>`

        :returns: The parsed AST representation of ``as_str`` and the updated string.
        :rtype: 2-member :class:`tuple <python:tuple>` of :class:`esprima.nodes.Script`
          and :class:`str <python:str>`
        """
        try:
            parsed = esprima.parseScript(as_str, loc = range, range = range)
        except ParseError:
            try:
                parsed = esprima.parseModule(as_str, loc = range, range = range)
            except ParseError:
                if not _break_loop_on_failure:
                    as_str = f"""var randomVariable = {as_str}"""
                    return cls._validate_js_literal(as_str,
                                                    range = range,
                                                    _break_loop_on_failure = True)
                else:
                    raise errors.HighchartsParseError('._validate_js_function() expects '
                                                      'a str containing a valid '
                                                      'JavaScript function. Could not '
                                                      'find a valid function.')

        return parsed, as_str



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from highcharts_core.options.accessibility import Accessibility as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal

STANDARD_PARAMS = [
    ({}, None),
    ({
      'announce_new_data': {
          'announcementFormatter': """function() { return true; }""",
          'enabled': True,
          'interruptUser': True,
          'minimumAnnouncementInterval': 3
      },
      'custom_components': {
          'item1': 'some-value-goes-here',
          'item2': 'some_other_value_goes_here'
      },
      'description': 'Description goes here',
      'enabled': True,
      'high_contrast_theme': 'something goes here',
      'keyboard_navigation': {
          'enabled': True,
          'focusBorder': {
              'enabled': True,
              'hideBrowserFocusOutline': False,
              'margin': 20,
              'style': {
                  'borderRadius': 2,
                  'color': '#ccc',
                  'lineWidth': 1
              }
          },
          'order': [
              'series',
              'item1',
              'zoom',
              'container'
          ],
          'seriesNavigation': {
              'mode': 'normal',
              'pointNavigationEnabledThreshold': 20,
              'rememberPointFocus': True,
              'skipNullPoints': False
          },
          'wrapAround': True
      },
      'landmark_verbosity': 'all',
      'linked_description': 'some-item-goes-here',
      'point': {
          'dateFormat': 'format string',
          'dateFormatter': """function() { return true; }""",
          'describeNull': False,
          'descriptionFormatter': """function() { return true; }""",
          'valueDecimals': 2,
          'valueDescriptionFormat': 'format string',
          'valuePrefix': '$',
          'valueSuffix': 'USD'
      },
      'screen_reader_section': {
          'afterChartFormat': 'format string',
          'afterChartFormatter': """function() { return true; }""",
          'axisRangeDateFormat': 'format string',
          'beforeChartFormat': 'format string',
          'beforeChartFormatter': """function() { return true; }""",
          'onPlayAsSoundClick': """function() { return true; }""",
          'onViewDataTableClick': """function() { return true; }"""
      },
      'series': {
          'describeSingleSeries': True,
          'descriptionFormat': 'format string goes here',
          'descriptionFormatter': """function() { return true; }""",
          'pointDescriptionEnabledThreshold': 100
      },
      'type_description': 'some description goes here'
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('accessibility/accessibility/01.js', False, None),

    ('accessibility/accessibility/error-01.js',
     False,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

    ('accessibility/accessibility/01.js', True, None),

    ('accessibility/accessibility/error-01.js',
     True,
     (errors.HighchartsValueError,
      errors.HighchartsParseError,
      JSONDecodeError,
      TypeError,
      ValueError)),

])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.animation import AnimationOptions


class LegendNavigation(HighchartsMeta):
    """Options for paging or navigation within the legend when the legend overflows
    the available space.

    .. hint::

      Navigation works well on screen, but not in static exported images. One way of
      working around that is to increase the chart height in export.

    """

    def __init__(self, **kwargs):
        self._active_color = None
        self._animation = None
        self._arrow_size = None
        self._enabled = None
        self._inactive_color = None
        self._style = None

        self.active_color = kwargs.get('active_color', None)
        self.animation = kwargs.get('animation', None)
        self.arrow_size = kwargs.get('arrow_size', None)
        self.enabled = kwargs.get('enabled', None)
        self.inactive_color = kwargs.get('inactive_color', None)
        self.style = kwargs.get('style', None)

    @property
    def active_color(self) -> Optional[str | Gradient | Pattern]:
        """The color or gradient of the active up or down arrow in the legend page
        navigation. Defaults to
        ``'#003399'``.

        :returns: The color of the active up or down arrow in the legend page navigation.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._active_color

    @active_color.setter
    def active_color(self, value):
        from highcharts_core import utility_functions
        self._active_color = utility_functions.validate_color(value)

    @property
    def animation(self) -> Optional[bool | AnimationOptions]:
        """Determines how to animate the legend items when navigating up or down.

        A value of ``True`` applies the default navigation given in the
        :meth:`Chart.animation`` setting. Additional options can be given as an object
        containing values for easing and duration.

        :returns: The configuration settings for the annotation animation.
        :rtype: :class:`AnnotationAnimation` or :class:`bool <python:bool>` or
          :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    def animation(self, value):
        if not value:
            self._animation = None
        elif value is True:
            self._animation = True
        else:
            self._animation = validate_types(value, types = AnimationOptions)

    @property
    def arrow_size(self) -> Optional[int | float | Decimal]:
        """The pixel size of the up and down arrows in the legend paging navigation.
        Defaults to ``12``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._arrow_size

    @arrow_size.setter
    def arrow_size(self, value):
        self._arrow_size = validators.numeric(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables keyboard navigation for the legend. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def inactive_color(self) -> Optional[str | Gradient | Pattern]:
        """The color or gradient of the inactive up or down arrow in the legend page
        navigation. Defaults to
        ``'#cccccc'``.

        :returns: The color of the active up or down arrow in the legend page navigation.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._inactive_color

    @inactive_color.setter
    def inactive_color(self, value):
        from highcharts_core import utility_functions
        self._inactive_color = utility_functions.validate_color(value)

    @property
    def style(self) -> Optional[str]:
        """CSS styling to apply to the legend page navigation.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'active_color': as_dict.get('activeColor', None),
            'animation': as_dict.get('animation', None),
            'arrow_size': as_dict.get('arrowSize', None),
            'enabled': as_dict.get('enabled', None),
            'inactive_color': as_dict.get('inactiveColor', None),
            'style': as_dict.get('style', None),
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'activeColor': self.active_color,
            'animation': self.animation,
            'arrowSize': self.arrow_size,
            'enabled': self.enabled,
            'inactiveColor': self.inactive_color,
            'style': self.style
        }
        return untrimmed



from typing import Optional

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AnnotationControlPointOption(HighchartsMeta):
    """Options for annotation's control points."""

    def __init__(self, **kwargs):
        self._positioner = None

        self.positioner = kwargs.get('positioner', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'annotations.controlPointOptions'

    @property
    def positioner(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to modify annotation's positioner controls.

        The JavaScript function should receive two arguments, ``this`` being the
        annotation's control point and ``target`` being the annotation's controllable.

        :returns: A JavaScript callback function to modify the annotation's positioner
          controls.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._positioner

    @positioner.setter
    @class_sensitive(CallbackFunction)
    def positioner(self, value):
        self._positioner = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'positioner': as_dict.get('positioner', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'positioner': self.positioner
        }

        return untrimmed



import datetime
from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.utility_classes.buttons import CollapseButtonConfiguration
from highcharts_core.utility_classes.events import SeriesEvents
from highcharts_core.utility_classes.javascript_functions import CallbackFunction
from highcharts_core.options.plot_options.link import LinkOptions


class TreegraphEvents(SeriesEvents):
    """General event handlers for the series items. 
    
    .. tip::
    
      These event hooks can also be attached to the series at run time using the ``Highcharts.addEvent()`` (JavaScript) 
      function.
      
    """
    
    def __init__(self, **kwargs):
        self._set_root_node = None
        
        self.set_root_node = kwargs.get('set_root_node', None)
        
        super().__init__(**kwargs)
        
    @property
    def set_root_node(self) -> Optional[CallbackFunction]:
        """Event handler that fires on a request to change the tree's root node, *before* the update is made. 
        
        An event object is passed to the function, containing additional properties ``newRootId``, ``previousRootId``, 
        ``redraw``, and ``trigger``.
        
        Defaults to :obj:`None <python:None>`
        
        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._set_root_node
    
    @set_root_node.setter
    @class_sensitive(CallbackFunction)
    def set_root_node(self, value):
        self._set_root_node = value
        
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'after_animate': as_dict.get('afterAnimate', None),
            'checkbox_click': as_dict.get('checkboxClick', None),
            'click': as_dict.get('click', None),
            'hide': as_dict.get('hide', None),
            'legend_item_click': as_dict.get('legendItemClick', None),
            'mouse_out': as_dict.get('mouseOut', None),
            'mouse_over': as_dict.get('mouseOver', None),
            'show': as_dict.get('show', None),
            
            'set_root_node': as_dict.get('setRootNode', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'afterAnimate': self.after_animate,
            'checkboxClick': self.checkbox_click,
            'click': self.click,
            'hide': self.hide,
            'legendItemClick': self.legend_item_click,
            'mouseOut': self.mouse_out,
            'mouseOver': self.mouse_over,
            'setRootNode': self.set_root_node,
            'show': self.show
        }

        return untrimmed


class TreegraphOptions(GenericTypeOptions):
    """General options to apply to all :term:`Treegraph` series types.
    
    A treegraph visualizes a relationship between ancestors and descendants with a clear parent-child relationship,
    e.g. a family tree or a directory structure.
    
    .. figure:: ../../../_static/treegraph-example.png
      :alt: Treegraph Example Chart
      :align: center
    
    """

    def __init__(self, **kwargs):
        self._animation_limit = None
        self._boost_blending = None
        self._boost_threshold = None
        self._color_index = None
        self._crisp = None
        self._crop_threshold = None
        self._find_nearest_point_by = None
        self._get_extremes_from_all = None
        self._relative_x_value = None
        self._soft_threshold = None
        self._step = None

        self._point_interval = None
        self._point_interval_unit = None
        self._point_start = None
        self._stacking = None

        self._allow_traversing_tree = None
        self._collapse_button = None
        self._color_by_point = None
        self._fill_space = None
        self._link = None
        self._reversed = None        
        self._traverse_up_button = None
        
        self.animation_limit = kwargs.get('animation_limit', None)
        self.boost_blending = kwargs.get('boost_blending', None)
        self.boost_threshold = kwargs.get('boost_threshold', None)
        self.color_index = kwargs.get('color_index', None)
        self.crisp = kwargs.get('crisp', None)
        self.crop_threshold = kwargs.get('crop_threshold', None)
        self.find_nearest_point_by = kwargs.get('find_nearest_point_by', None)
        self.get_extremes_from_all = kwargs.get('get_extremes_from_all', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)
        self.soft_threshold = kwargs.get('soft_threshold',  None)
        self.step = kwargs.get('step', None)

        self.point_interval = kwargs.get('point_interval', None)
        self.point_interval_unit = kwargs.get('point_interval_unit', None)
        self.point_start = kwargs.get('point_start', None)
        self.stacking = kwargs.get('stacking', None)

        self.allow_traversing_tree = kwargs.get('allow_traversing_tree', None)
        self.collapse_button = kwargs.get('collapse_button', None)
        self.color_by_point = kwargs.get('color_by_point', None)
        self.fill_space = kwargs.get('fill_space', None)
        self.link = kwargs.get('link', None)
        self.reversed = kwargs.get('reversed', None)
        
        super().__init__(**kwargs)
        
    @property
    def animation_limit(self) -> Optional[int | float | Decimal]:
        """For some series, there is a limit that shuts down initial animation by default
        when the total number of points in the chart is too high. Defaults to
        :obj:`None <python:None>`.

        For example, for a column chart and its derivatives, animation does not run if
        there is more than 250 points totally. To disable this cap, set
        ``animation_limit`` to ``float("inf")`` (which represents infinity).

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._animation_limit

    @animation_limit.setter
    def animation_limit(self, value):
        if value == float('inf'):
            self._animation_limit = float('inf')
        else:
            self._animation_limit = validators.numeric(value,
                                                       allow_empty = True,
                                                       minimum = 0)

    @property
    def boost_blending(self) -> Optional[str]:
        """Sets the color blending in the boost module. Defaults to
        :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._boost_blending

    @boost_blending.setter
    def boost_blending(self, value):
        self._boost_blending = validators.string(value, allow_empty = True)

    @property
    def boost_threshold(self) -> Optional[int]:
        """Set the point threshold for when a series should enter boost mode. Defaults to
        ``5000``.

        Setting it to e.g. 2000 will cause the series to enter boost mode when there are
        2,000 or more points in the series.

        To disable boosting on the series, set the ``boost_threshold`` to ``0``. Setting
        it to ``1`` will force boosting.

        .. note::

          The :meth:`AreaOptions.crop_threshold` also affects this setting.

          When zooming in on a series that has fewer points than the ``crop_threshold``,
          all points are rendered although outside the visible plot area, and the
          ``boost_threshold`` won't take effect.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._boost_threshold

    @boost_threshold.setter
    def boost_threshold(self, value):
        self._boost_threshold = validators.integer(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        .. tip::
        
          .. versionadded:: Highcharts (JS) v.11

          With Highcharts (JS) v.11, using CSS variables of the form ``--highcharts-color-{n}`` make
          changing the color scheme very simple.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def crisp(self) -> Optional[bool]:
        """If ``True``, each point or column edge is rounded to its nearest pixel in order
        to render sharp on screen. Defaults to ``True``.

        .. hint::

          In some cases, when there are a lot of densely packed columns, this leads to
          visible difference in column widths or distance between columns. In these cases,
          setting ``crisp`` to ``False`` may look better, even though each column is
          rendered blurry.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crisp

    @crisp.setter
    def crisp(self, value):
        if value is None:
            self._crisp = None
        else:
            self._crisp = bool(value)

    @property
    def crop_threshold(self) -> Optional[int]:
        """When the series contains less points than the crop threshold, all points are
        drawn, even if the points fall outside the visible plot area at the current zoom.
        Defaults to ``300``.

        The advantage of drawing all points (including markers and columns), is that
        animation is performed on updates. On the other hand, when the series contains
        more points than the crop threshold, the series data is cropped to only contain
        points that fall within the plot area. The advantage of cropping away invisible
        points is to increase performance on large series.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._crop_threshold

    @crop_threshold.setter
    def crop_threshold(self, value):
        self._crop_threshold = validators.integer(value,
                                                   allow_empty = True,
                                                   minimum = 0)

    @property
    def events(self) -> Optional[TreegraphEvents]:
        """General event handlers for the series items.

        .. note::

          These event hooks can also be attached to the series at run time using the
          (JavaScript) ``Highcharts.addEvent()`` function.

        :rtype: :class:`TreegraphEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(TreegraphEvents)
    def events(self, value):
        self._events = value

    @property
    def find_nearest_point_by(self) -> Optional[str]:
        """Determines whether the series should look for the nearest point in both
        dimensions or just the x-dimension when hovering the series.

        If :obj:`None <python:None>`, defaults to ``'xy'`` for scatter series and ``'x'``
        for most other series. If the data has duplicate x-values, it is recommended to
        set this to ``'xy'`` to allow hovering over all points.

        Applies only to series types using nearest neighbor search (not direct hover) for
        tooltip.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._find_nearest_point_by

    @find_nearest_point_by.setter
    def find_nearest_point_by(self, value):
        self._find_nearest_point_by = validators.string(value, allow_empty = True)

    @property
    def get_extremes_from_all(self) -> Optional[bool]:
        """If ``True``, uses the Y extremes of the total chart width or only the zoomed
        area when zooming in on parts of the X axis. By default, the Y axis adjusts to the
        min and max of the visible data.

        .. warning::

          Applies to :term:`Cartesian series` only.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._get_extremes_from_all

    @get_extremes_from_all.setter
    def get_extremes_from_all(self, value):
        if value is None:
            self._get_extremes_from_all = None
        else:
            self._get_extremes_from_all = bool(value)

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @property
    def soft_threshold(self) -> Optional[bool]:
        """When ``True``, the series will not cause the Y axis to cross the zero plane (or
        threshold option) unless the data actually crosses the plane. Defaults to
        ``True``.

        For example, if ``False``, a series of ``0, 1, 2, 3`` will make the Y axis show
        negative values according to the ``min_padidng`` option. If ``True``, the Y axis
        starts at 0.

        :rtype: :class:`bool <python:bool>`
        """
        return self._soft_threshold

    @soft_threshold.setter
    def soft_threshold(self, value):
        if value is None:
            self._soft_threshold = None
        else:
            self._soft_threshold = bool(value)

    @property
    def step(self) -> Optional[str]:
        """Whether to apply steps to the line. Defaults to :obj:`None <python:None>`.

        Possible values are:

          * :obj:`None <python:None>`
          * ``'left'``
          * ``'center'``
          * ``'right'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._step

    @step.setter
    def step(self, value):
        self._step = validators.string(value, allow_empty = True)

    @property
    def point_interval(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_interval`` defines
        the interval of the x values. Defaults to ``1``.

        For example, if a series contains one value every decade starting from year 0, set
        ``point_interval`` to ``10``. In true datetime axes, the ``point_interval`` is set
        in milliseconds.

        .. hint::

          ``point_interval`` can be also be combined with
          :meth:`point_interval_unit <AreaOptions.point_interval_unit>` to draw irregular
          time intervals.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point, and the ``point_interval`` is added x times to
          the :meth:`point_start <AreaOptions.point_start>` setting.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_interval

    @point_interval.setter
    def point_interval(self, value):
        self._point_interval = validators.numeric(value,
                                                  allow_empty = True,
                                                  minimum = 0)

    @property
    def point_interval_unit(self) -> Optional[str]:
        """On datetime series, this allows for setting the
        :meth:`point_interval <AreaOptions.point_interval>` to irregular time units, day,
        month, and year.

        A day is usually the same as 24 hours, but ``point_interval_unit`` also takes the
        DST crossover into consideration when dealing with local time.

        Combine this option with :meth:`point_interval <AreaOptions.point_interval>` to
        draw weeks, quarters, 6 month periods, 10 year periods, etc.

        .. warning::

          This options applies to the series data, not the interval of the axis ticks,
          which is independent.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._point_interval_unit

    @point_interval_unit.setter
    def point_interval_unit(self, value):
        self._point_interval_unit = validators.string(value, allow_empty = True)

    @property
    def point_start(self) -> Optional[int | float | Decimal]:
        """If no x values are given for the points in a series, ``point_start`` defines
        on what value to start. For example, if a series contains one yearly value
        starting from 1945, set ``point_start`` to ``1945``. Defaults to ``0``.

        .. note::

          If combined with :meth:`relative_x_value <AreaOptions.relative_x_value>`, an x
          value can be set on each point. The x value from the point options is multiplied
          by :meth:`point_interval <AreaOptions.point_interval>` and added to
          ``point_start`` to produce a modified x value.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._point_start

    @point_start.setter
    def point_start(self, value):
        try:
            value = validators.numeric(value, allow_empty = True)
        except (TypeError, ValueError) as error:
            value = validators.datetime(value)

            if hasattr(value, 'timestamp') and value.tzinfo is not None:
                self._point_start = value.timestamp()*1000
            elif hasattr(value, 'timestamp'):
                value = value.replace(tzinfo = datetime.timezone.utc)
                value = value.timestamp()*1000
            else:
                raise error
            
        self._point_start = value

    @property
    def stacking(self) -> Optional[str]:
        """Whether to stack the values of each series on top of each other. Defaults to
        :obj:`None <python:None>`.

        Acceptable values are:

          * :obj:`None <python:None>` to disable stacking,
          * ``"normal"`` to stack by value or
          * ``"percent"``
          * ``'stream'`` (for streamgraph series type only)
          * ``'overlap'`` (for waterfall series type only)

        .. note::

          When stacking is enabled, data must be sorted in ascending X order.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stacking

    @stacking.setter
    def stacking(self, value):
        if not value:
            self._stacking = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['normal', 'percent', 'stream', 'overlap']:
                raise errors.HighchartsValueError(f'stacking expects a valid stacking '
                                                  f'value. However, received: {value}')
            self._stacking = value

    @property
    def allow_traversing_tree(self) -> Optional[bool]:
        """If ``True``, the user can click on a point which is a parent and zoom in on its
        children. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_traversing_tree

    @allow_traversing_tree.setter
    def allow_traversing_tree(self, value):
        if value is None:
            self._allow_traversing_tree = None
        else:
            self._allow_traversing_tree = bool(value)

    @property
    def collapse_button(self) -> Optional[CollapseButtonConfiguration]:
        """Options applied to the Collapse Button, which is the small button that indicates the node is collapsible.
        
        :rtype: :class:`CollapseButtonConfiguration <highcharts_core.utility_classes.buttons.CollapseButtonConfiguration>` 
          or :obj:`None <python:None>`
        """
        return self._collapse_button
    
    @collapse_button.setter
    @class_sensitive(CollapseButtonConfiguration)
    def collapse_button(self, value):
        self._collapse_button = value
        
    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def fill_space(self) -> Optional[bool]:
        """If ``True``, the treegraph series should fill the entire plot area in the 
        X-axis direction, even when there are collapsed points. Defaults to ``False``.
        
        :rtype: :class:`bool <python:bool>`
        """
        return self._fill_space
    
    @fill_space.setter
    def fill_space(self, value):
        if value is None:
            self._fill_space = None
        else:
            self._fill_space = bool(value)
            
    @property
    def link(self) -> Optional[LinkOptions]:
        """Link style options.

        :rtype: :class:`LinkOptions` or :obj:`None <python:None>`
        """
        return self._link

    @link.setter
    @class_sensitive(LinkOptions)
    def link(self, value):
        self._link = value

    @property
    def reversed(self) -> Optional[bool]:
        """If ``True``, places the series on the other side of the plot area. Defaults to
        ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._reversed

    @reversed.setter
    def reversed(self, value):
        if value is None:
            self._reversed = None
        else:
            self._reversed = bool(value)
            
    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_index': as_dict.get('colorIndex', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'step': as_dict.get('step', None),

            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_start': as_dict.get('pointStart', None),
            'stacking': as_dict.get('stacking', None),

            'allow_traversing_tree': as_dict.get('allowTraversingTree', None),
            'collapse_button': as_dict.get('collapseButton', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'fill_space': as_dict.get('fillSpace', None),
            'link': as_dict.get('link', None),
            'reversed': as_dict.get('reversed', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'allowPointSelect': self.allow_point_select,
            'animation': self.animation,
            'className': self.class_name,
            'clip': self.clip,
            'color': self.color,
            'cursor': self.cursor,
            'custom': self.custom,
            'dashStyle': self.dash_style,
            'dataLabels': self.data_labels,
            'description': self.description,
            'enableMouseTracking': self.enable_mouse_tracking,
            'events': self.events,
            'includeInDataExport': self.include_in_data_export,
            'keys': self.keys,
            'label': self.label,
            'linkedTo': self.linked_to,
            'marker': self.marker,
            'onPoint': self.on_point,
            'opacity': self.opacity,
            'point': self.point,
            'pointDescriptionFormatter': self.point_description_formatter,
            'selected': self.selected,
            'showCheckbox': self.show_checkbox,
            'showInLegend': self.show_in_legend,
            'skipKeyboardNavigation': self.skip_keyboard_navigation,
            'states': self.states,
            'stickyTracking': self.sticky_tracking,
            'threshold': self.threshold,
            'tooltip': self.tooltip,
            'turboThreshold': self.turbo_threshold,
            'visible': self.visible,
            'type': self.type,

            'animationLimit': self.animation_limit,
            'boostBlending': self.boost_blending,
            'boostThreshold': self.boost_threshold,
            'colorIndex': self.color_index,
            'crisp': self.crisp,
            'cropThreshold': self.crop_threshold,
            'findNearestPointBy': self.find_nearest_point_by,
            'getExtremesFromAll': self.get_extremes_from_all,
            'relativeXValue': self.relative_x_value,
            'softThreshold': self.soft_threshold,
            'step': self.step,

            'pointInterval': self.point_interval,
            'pointIntervalUnit': self.point_interval_unit,
            'pointStart': self.point_start,
            'stacking': self.stacking,

            'allowTraversingTree': self.allow_traversing_tree,
            'collapseButton': self.collapse_button,
            'colorByPoint': self.color_by_point,
            'fillSpace': self.fill_space,
            'link': self.link,
            'reversed': self.reversed,
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.date_time_label_formats import DateTimeLabelFormats
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Tooltip(HighchartsMeta):
    """Options for the tooltip that appears when the user hovers over a series or
    point."""

    def __init__(self, **kwargs):
        self._animation = None
        self._background_color = None
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._class_name = None
        self._cluster_format = None
        self._date_time_label_formats = None
        self._distance = None
        self._enabled = None
        self._follow_pointer = None
        self._follow_touch_move = None
        self._footer_format = None
        self._format = None
        self._formatter = None
        self._header_format = None
        self._header_shape = None
        self._hide_delay = None
        self._null_format = None
        self._null_formatter = None
        self._outside = None
        self._padding = None
        self._point_format = None
        self._point_formatter = None
        self._positioner = None
        self._shadow = None
        self._shape = None
        self._shared = None
        self._snap = None
        self._split = None
        self._stick_on_contact = None
        self._style = None
        self._use_html = None
        self._value_decimals = None
        self._value_prefix = None
        self._value_suffix = None
        self._x_date_format = None

        self.animation = kwargs.get('animation', None)
        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.cluster_format = kwargs.get('cluster_format', None)
        self.date_time_label_formats = kwargs.get('date_time_label_formats', None)
        self.distance = kwargs.get('distance', None)
        self.enabled = kwargs.get('enabled', None)
        self.follow_pointer = kwargs.get('follow_pointer', None)
        self.follow_touch_move = kwargs.get('follow_touch_move', None)
        self.footer_format = kwargs.get('footer_format', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.header_format = kwargs.get('header_format', None)
        self.header_shape = kwargs.get('header_shape', None)
        self.hide_delay = kwargs.get('hide_delay', None)
        self.null_format = kwargs.get('null_format', None)
        self.null_formatter = kwargs.get('null_formatter', None)
        self.outside = kwargs.get('outside', None)
        self.padding = kwargs.get('padding', None)
        self.point_format = kwargs.get('point_format', None)
        self.point_formatter = kwargs.get('point_formatter', None)
        self.positioner = kwargs.get('positioner', None)
        self.shadow = kwargs.get('shadow', None)
        self.shape = kwargs.get('shape', None)
        self.shared = kwargs.get('shared', None)
        self.snap = kwargs.get('snap', None)
        self.split = kwargs.get('split', None)
        self.stick_on_contact = kwargs.get('stick_on_contact', None)
        self.style = kwargs.get('style', None)
        self.use_html = kwargs.get('use_html', None)
        self.value_decimals = kwargs.get('value_decimals', None)
        self.value_prefix = kwargs.get('value_prefix', None)
        self.value_suffix = kwargs.get('value_suffix', None)
        self.x_date_format = kwargs.get('x_date_format', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'tooltip'

    @property
    def animation(self) -> Optional[bool]:
        """Flag which indicates whether animation is enabled on the toltip (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    def animation(self, value):
        if value is None:
            self._animation = None
        else:
            self._animation = bool(value)

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the tooltip. Defaults to
        ``None``.

        :returns: The backgorund color for the tooltip.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the tooltip border.

        :returns: The color of the tooltip's border.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`

        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        from highcharts_core import utility_functions
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_radius(self) -> Optional[int | float | Decimal | str]:
        """The border radius (in pixels) applied to
        the pane. Defaults to ``3``.

        :returns: The border radius of the tooltip.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        if value is None or value == '':
            self._border_radius = None
        else:
            try:
                self._border_radius = validators.string(value, allow_empty = True)
            except (TypeError, ValueError):
                self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the tooltip. Defaults to
        ``1``.

        :returns: The border width to apply to the tooltip.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS. Defaults to :obj:`None <python:None>`.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def cluster_format(self) -> Optional[str]:
        """The HTML of the cluster point's in the tooltip.

        .. warning::

          Works only with ``marker-clusters`` module and analogously to
          :meth:`Tooltip.point_format`.

        .. note::

          The cluster tooltip can be also formatted using the :meth:`Tooltip.formatter`
          callback function and the :meth:`Point.is_cluster` flag.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._cluster_format

    @cluster_format.setter
    def cluster_format(self, value):
        self._cluster_format = validators.string(value, allow_empty = True)

    @property
    def date_time_label_formats(self) -> Optional[DateTimeLabelFormats]:
        """For series on datetime axes, the date format in the tooltip's header will by
        default be guessed based on the closest data points. This property gives the
        default string representations used for each unit.

        Defaults to :obj:`None <python:None>`

        :rtype: :class:`DateTimeLabelFormats` or :obj:`None <python:None>`
        """
        return self._date_time_label_formats

    @date_time_label_formats.setter
    @class_sensitive(DateTimeLabelFormats)
    def date_time_label_formats(self, value):
        self._date_time_label_formats = value

    @property
    def distance(self) -> Optional[int | float | Decimal]:
        """The distance (in pixels) from the point to the tooltip. Defaults to
        ``16``.

        :returns: The distance from the point to the tooltip.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._distance

    @distance.setter
    def distance(self, value):
        self._distance = validators.numeric(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables the use of tooltips. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def follow_pointer(self) -> Optional[bool]:
        """If ``True``, the tooltip will follow the mouse pointer as it moves across
        columns, pie slices, and other point types with an extent. Defaults to ``False``
        generally, except for pie, polygon, map, sankey, and wordcloud series types where
        the generic ``False`` is over-ridden by default.

        .. note::

          If :meth:`Tooltip.split` is ``True``, then this property is ignored.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._follow_pointer

    @follow_pointer.setter
    def follow_pointer(self, value):
        if value is None:
            self._follow_pointer = None
        else:
            self._follow_pointer = bool(value)

    @property
    def follow_touch_move(self) -> Optional[bool]:
        """If ``True``, the tooltip will follow the single finger touches on a touch base.
        Defaults to ``True``.

        .. note::

          If this is ``True`` and :meth:`Chart.panning` is set, this property will take
          over one-finger touch moves so the user will need to use two fingers for zooming
          and panning.

        .. note::

          There is a significant difference to behavior when compared to
          :meth:`Tooltip.follow_pointer`, which updates the *position* of the tooltip. For
          example, if :meth:`follow_pointer <Tooltip.follow_pointer>` is ``False`` for a
          column series, the tooltip will show above or below the column. However, when
          ``follow_touch_move`` is ``True``, the tooltip displayed will jump from column
          to column (above or below the column as applicable) as the user swipes across
          the plot area.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._follow_touch_move

    @follow_touch_move.setter
    def follow_touch_move(self, value):
        if value is None:
            self._follow_touch_move = None
        else:
            self._follow_touch_move = bool(value)

    @property
    def footer_format(self) -> Optional[str]:
        """A string to append to the tooltip format. Defaults to
        ``''`` (an empty string).

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._footer_format

    @footer_format.setter
    def footer_format(self, value):
        if value == '':
            self._footer_format = value
        else:
            self._footer_format = validators.string(value, allow_empty = True)

    @property
    def format(self) -> Optional[str]:
        """A :term:`format string` for the entire tooltip. Defaults to :obj:`None <python:None>`.

        .. tip::

          When format strings are required, it is usually more convenient to use
          :meth:`.header_format <highcharts_core.options.tooltips.Tooltip.header_format>`,
          :meth:`.point_format <highcharts_core.options.tooltips.Tooltip.point_format>`, and
          :meth:`.footer_format <highcharts_core.options.tooltips.Tooltip.footer_format>`.
          
          However, the ``.format`` option allows combining them into one setting.
          
        .. note::
        
          The context of the format string is the same as that of the
          :meth:`.formatter <highcharts_core.options.tooltips.Tooltip.formatter>` callback.
        
        :rtype: :class:`str <python:str>`  
        """
        return self._format
    
    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the text of the tooltip from scratch.
        Defaults to :obj:`None <python:None>`

        In case of single or :meth:`shared <Tooltip.shared>`tooltips, a string should be
        returned. In case of :meth:`split <Tooltip.split>` tooltips, it should return an
        array where the first item is the header, and subsequent items are mapped to the
        points. Return ``false`` to disable the tooltip for a specific point on a series.

        .. note::

          A subset of HTML is supported. Unless :meth:`Tooltip.use_html` is ``True``, the
          HTML of the tooltip is parsed and converted to SVG, therefore this is *not* a
          complete HTML renderer. The following HTML tags are supported:
          ``b``, ``br``, ``em``, ``i``, ``span``, ``strong``.

          Spans can be styled with a ``style`` attribute, but only text-related CSS that
          is shared with SVG will be handled.

        .. note::

          The available data in the formatter differ a bit depending on whether the
          tooltip is shared, split, or belongs to a single point. In a shared/split
          tooltip, all properties except ``x``, which is common for all points, are kept
          in an array, ``this.points``.

          Available data are:

            * ``this.percentage`` (when not shared) / ``this.points[i].percentage``
              (when shared): Stacked series and pies only. The point's percentage of the
              total.
            * ``this.point`` (when not shared) / ``this.points[i].point`` (when shared):
              The point object. The point name, if defined, is available through
              ``this.point.name``.
            * ``this.points``: In a shared tooltip, this is an array containing all other
              properties for each point.
            * ``this.series`` (when not shared) / ``this.points[i].series`` (when shared):
              The series object. The series name is available through
              ``this.series.name``.
            * ``this.total`` (when not shared) / ``this.points[i].total`` (when shared):
              Stacked series only. The total value at this point's x value.
            * ``this.x``: The x value. This property is the same regardless of the tooltip
              being shared or not.
            * ``this.y`` (when not shared) / ``this.points[i].y`` (when shared): The y
              value.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def header_format(self) -> Optional[str]:
        """The HTML of the tooltip header line. Defaults to :obj:`None <python:None>`.

        .. note::

          Variables are enclosed by curly brackets. Available variables are ``point.key``,
          ``series.name``, ``series.color``, and other members from the ``point`` and
          ``series`` objects.

          The ``point.key`` variable contains the category name, x value or datetime
          string depending on the type of axis. For datetime axes, the ``point.key`` date
          format can be set using :meth:`ToolTip.x_date_format`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._header_format

    @header_format.setter
    def header_format(self, value):
        self._header_format = validators.string(value, allow_empty = True)

    @property
    def header_shape(self) -> Optional[str]:
        """The name of a symbol to use for the border around the tooltip header.

        Defaults to ``'callout'``.

        .. note::

          Applies only when :meth:`Tooltip.split` is enabled.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._header_shape

    @header_shape.setter
    def header_shape(self, value):
        if not value:
            self._header_shape = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['callout', 'circle', 'square']:
                raise errors.HighchartsValueError(f'shape expects a supported tooltip '
                                                  f'header shape. Was: {value}')
            self._header_shape = value

    @property
    def hide_delay(self) -> Optional[int]:
        """The number of milliseconds to wait until the tooltip is hidden when mouse out
        from a point or chart.

        Defaults to ``{constants.DEFAULT_TOOLTIP.get('hide_delay')}``.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._hide_delay

    @hide_delay.setter
    def hide_delay(self, value):
        self._hide_delay = validators.integer(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def null_format(self) -> Optional[str]:
        """The HTML of the null point's line in the tooltip. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Works analogously to :meth:`Tooltip.point_format`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._null_format

    @null_format.setter
    def null_format(self, value):
        self._null_format = validators.string(value, allow_empty = True)

    @property
    def null_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the text of the tooltip for visible null
        points.

        .. note::

          Works analogously to :meth:`Tooltip.formatter`.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._null_formatter

    @null_formatter.setter
    @class_sensitive(CallbackFunction)
    def null_formatter(self, value):
        self._null_formatter = value

    @property
    def outside(self) -> Optional[bool]:
        """If ``True``, allows the tooltip to render outside of the chart's SVG element
        box. Defaults to ``False``, which causes the tooltip to be rendered inside the
        SVG element box which results in the tooltip being rendered inside the chart area.


        .. warning::

          When setting this property to ``False`` (the default) on small charts, the
          tooltip may either clip or overlap parts of the chart itself.

        When ``True``, a separate SVG element is created and overlaid on the page,
        allowing the tooltip to be aligned inside the page itself.

        .. note::

          If :meth:`Chart.scrollable_plot_area` is ``True``, then this will default to
          ``True``. Otherwise, it will default to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._outside

    @outside.setter
    def outside(self, value):
        if value is None:
            self._outside = None
        else:
            self._outside = bool(value)

    @property
    def padding(self) -> Optional[int | float | Decimal]:
        """The padding inside the tooltip, expressed in pixels.

        Defaults to ``8``.

        :returns: The padding to apply to the tooltip.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validators.numeric(value, allow_empty = True)

    @property
    def point_format(self) -> Optional[str]:
        """The HTML of the point's line in the tooltip. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Variables are enclosed by curly brackets. Available variables are ``point.x``,
          ``point.y``, ``series.name``, ``series.color``, and other properties of the same
          form.

          Furthermore, ``point.y`` can be extended by the :meth:`Tooltip.value_prefix` and
          :meth:`Tooltip.value_suffix` properties. This can also be overridden for each
          series, which makes it a good hook for displaying units.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._point_format

    @point_format.setter
    def point_format(self, value):
        self._point_format = validators.string(value, allow_empty = True)

    @property
    def point_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the text of the tooltip's point line.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._point_formatter

    @point_formatter.setter
    @class_sensitive(CallbackFunction)
    def point_formatter(self, value):
        self._point_formatter = value

    @property
    def positioner(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function to place the tooltip in a custom position.

        The callback receives three (JavaScript) parameters: ``labelWidth``,
        ``labelHeight``, and ``point``, where ``point`` contains values for ``plotX`` and
        ``plotY`` telling where the reference point is in the plot area. Add
        ``chart.plotLeft`` and ``chart.plotTop`` to get the full coordinates.

        To find the actual hovered ``Point`` instance, use ``this.chart.hoverPoint``. For
        :meth:`shared <Tooltip.shared>` or :meth:`split <Tooltip.split>` tooltips, all the
        hover points are available in ``this.chart.hoverPoints``.

        The return should be an object containing x and y values, for example:
        ``{ x: 100, y: 100 }``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._positioner

    @positioner.setter
    @class_sensitive(CallbackFunction)
    def positioner(self, value):
        self._positioner = value

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the tooltip. Defaults to
        ``True``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or a boolean setting which hides the
          shadow or displays the default shadow.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions` or
          :obj:`None <python:None>`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if value is None:
            self._shadow = None
        elif isinstance(value, bool) and value is False:
            self._shadow = False
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def shape(self) -> Optional[str]:
        """The name of the symbol to use for the border around the tooltip. Defaults to
        ``'callout'``.

        Accepts:

          * ``'rect'``
          * ``'circle'``
          * ``'callout'``

        .. note::

          When :meth:`Tooltip.split` is enabled, the shape is applied to all boxes except
          the header (which is controlled by :meth:`Tooltip.header_shape`).

        :returns: The shape to use for the border around the tooltip.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._shape

    @shape.setter
    def shape(self, value):
        if not value:
            self._shape = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['callout',
                             'rect',
                             'circle']:
                raise errors.HighchartsValueError(f'shape expects a supported tooltip '
                                                  f'shape. Was: {value}')
            self._shape = value

    @property
    def shared(self) -> Optional[bool]:
        """When ``True``, the entire plot area will capture mouse movement or touch
        events. Defaults to ``False``.

        .. hint::

          If ``True``, tooltip texts for series types with ordered data (not pie, scatter,
          flags etc) will be shown in a single bubble. This is recommended for single
          series charts and for tablet/mobile optimized charts.

          See also :meth:`Tooltip.split`, which is better suited for charts with many
          series, especially line-type series. The :meth:`Tooltip.split` option takes
          precedence over :meth:`Tooltip.shared`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._shared

    @shared.setter
    def shared(self, value):
        if value is None:
            self._shared = None
        else:
            self._shared = bool(value)

    @property
    def snap(self) -> Optional[int | float | Decimal]:
        """Proximity snap for graphs or single points. If :obj:`None <python:None>`, it
        defaults to ``10`` pixels for mouse-powered devices and ``25`` for touch devices.

        Defaults to :obj:`None <python:None>`.

        .. note::

          In most cases, the whole plot area captures the mouse movement, and in these
          cases :meth:`Tooltip.snap` doesn't make sense. This applies when
          :meth:`Tooltip.sticky_tracking` is ``True`` (default) and when the tooltip is
          :meth:`shared <Tooltip.shared>` or :meth:`split <Tooltip.split>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._snap

    @snap.setter
    def snap(self, value):
        self._snap = validators.numeric(value,
                                        allow_empty = True,
                                        minimum = 0)

    @property
    def split(self) -> Optional[bool]:
        """If ``True``, splits the tooltip into one label per series, with the header
        close to the axis. Defaults to ``False``.

        .. hint::

          This is recommended over :meth:`shared <Tooltip.shared>` tooltips for charts
          with multiple line series, generally making them easier to read.

        .. note::

          This option takes precedence over :meth:`Tooltip.shared`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._split

    @split.setter
    def split(self, value):
        if value is None:
            self._split = None
        else:
            self._split = bool(value)

    @property
    def stick_on_contact(self) -> Optional[bool]:
        """If ``True``, prevents the tooltip from switching or closing when touched or
        pointed. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._stick_on_contact

    @stick_on_contact.setter
    def stick_on_contact(self, value):
        if value is None:
            self._stick_on_contact = None
        else:
            self._stick_on_contact = bool(value)

    @property
    def style(self) -> Optional[str]:
        """CSS styling to apply to the tooltip.

        .. note::

          The tooltip can also be styled through the CSS class ``.highcharts-tooltip``.

        .. warning::

          The default ``pointerEvents`` style makes the tooltip ignore mouse events, so in
          order to use clickable tooltips, this value must be set to ``auto``.

        :rtype: :class:`str` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        self._style = validators.string(value, allow_empty = True, coerce_value = True)

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the tooltip. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        .. hint::

          Using HTML allows advanced formatting like tables and images in the tooltip. It
          is also recommended for RTL languages as it works around RTL bugs in early
          Firefox.

        :returns: Flag indicating whether to render tooltips using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def value_decimals(self) -> Optional[int]:
        """How many decimals to show in each series' y value. Defaults to
        :obj:`None <python:None>`, which perserves all decimals.

        .. note::

          This is overridable in each series' tooltip options object.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._value_decimals

    @value_decimals.setter
    def value_decimals(self, value):
        self._value_decimals = validators.integer(value, allow_empty = True)

    @property
    def value_prefix(self) -> Optional[str]:
        """A string to prepend to each series' y value. Overridable in each series'
        tooltip options object. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_prefix

    @value_prefix.setter
    def value_prefix(self, value):
        self._value_prefix = validators.string(value, allow_empty = True)

    @property
    def value_suffix(self) -> Optional[str]:
        """A string to append to each series' y value. Overridable in each series' tooltip
        options object. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_suffix

    @value_suffix.setter
    def value_suffix(self, value):
        self._value_suffix = validators.string(value, allow_empty = True)

    @property
    def x_date_format(self) -> Optional[str]:
        """The format for the date in the tooltip header if the X axis is a datetime axis.
        Defaults to :obj:`None <python:None>`, which produces a best-guess based on the
        smallest distance between points in the chart.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._x_date_format

    @x_date_format.setter
    def x_date_format(self, value):
        self._x_date_format = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'cluster_format': as_dict.get('clusterFormat', None),
            'date_time_label_formats': as_dict.get('dateTimeLabelFormats', None),
            'distance': as_dict.get('distance', None),
            'enabled': as_dict.get('enabled', None),
            'follow_pointer': as_dict.get('followPointer', None),
            'follow_touch_move': as_dict.get('followTouchMove', None),
            'footer_format': as_dict.get('footerFormat', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'header_format': as_dict.get('headerFormat', None),
            'header_shape': as_dict.get('headerShape', None),
            'hide_delay': as_dict.get('hideDelay', None),
            'null_format': as_dict.get('nullFormat', None),
            'null_formatter': as_dict.get('nullFormatter', None),
            'outside': as_dict.get('outside', None),
            'padding': as_dict.get('padding', None),
            'point_format': as_dict.get('pointFormat', None),
            'point_formatter': as_dict.get('pointFormatter', None),
            'positioner': as_dict.get('positioner', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'shared': as_dict.get('shared', None),
            'snap': as_dict.get('snap', None),
            'split': as_dict.get('split', None),
            'stick_on_contact': as_dict.get('stickOnContact', None),
            'style': as_dict.get('style', None),
            'use_html': as_dict.get('useHTML', None),
            'value_decimals': as_dict.get('valueDecimals', None),
            'value_prefix': as_dict.get('valuePrefix', None),
            'value_suffix': as_dict.get('valueSuffix', None),
            'x_date_format': as_dict.get('xDateFormat', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'animation': self.animation,
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'clusterFormat': self.cluster_format,
            'dateTimeLabelFormats': self.date_time_label_formats,
            'distance': self.distance,
            'enabled': self.enabled,
            'followPointer': self.follow_pointer,
            'followTouchMove': self.follow_touch_move,
            'footerFormat': self.footer_format,
            'format': self.format,
            'formatter': self.formatter,
            'headerFormat': self.header_format,
            'headerShape': self.header_shape,
            'hideDelay': self.hide_delay,
            'nullFormat': self.null_format,
            'nullFormatter': self.null_formatter,
            'outside': self.outside,
            'padding': self.padding,
            'pointFormat': self.point_format,
            'pointFormatter': self.point_formatter,
            'positioner': self.positioner,
            'shadow': self.shadow,
            'shape': self.shape,
            'shared': self.shared,
            'snap': self.snap,
            'split': self.split,
            'stickOnContact': self.stick_on_contact,
            'style': self.style,
            'useHTML': self.use_html,
            'valueDecimals': self.value_decimals,
            'valuePrefix': self.value_prefix,
            'valueSuffix': self.value_suffix,
            'xDateFormat': self.x_date_format
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class AccessibilityPoint(HighchartsMeta):
    """Options for describing individual points."""

    def __init__(self, **kwargs):
        self._date_format = None
        self._date_formatter = None
        self._describe_null = None
        self._description_format = None
        self._description_formatter = None
        self._value_decimals = None
        self._value_description_format = None
        self._value_prefix = None
        self._value_suffix = None

        self.date_format = kwargs.get('date_format', None)
        self.date_formatter = kwargs.get('date_formatter', None)
        self.describe_null = kwargs.get('describe_null', None)
        self.description_format = kwargs.get('description_format', None)
        self.description_formatter = kwargs.get('description_formatter', None)
        self.value_decimals = kwargs.get('value_decimals', None)
        self.value_description_format = kwargs.get('value_description_format', None)
        self.value_prefix = kwargs.get('value_prefix', None)
        self.value_suffix = kwargs.get('value_suffix', None)

    @property
    def date_format(self) -> Optional[str]:
        """Date format to use for points on datetime axes when describing them to screen
        reader users.

        Defaults to the same format as in tooltip.

        .. seealso::

          * Detailed documentation on supported format replacement codes:
            https://api.highcharts.com/class-reference/Highcharts.Time#dateFormat

        :returns: The date format to use for points on datetime axes when describing them
          to screen reader users.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._date_format

    @date_format.setter
    def date_format(self, value):
        self._date_format = validators.string(value, allow_empty = True)

    @property
    def date_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript formatter function to determine the date/time format used with
        points on datetime axes when describing them to screen reader users.

        The formatter function should receive one argument, ``point``, referring to the
        point to describe. Should return a date format string compatible with
        :meth:`AccessibilityPoint.date_format`.

        :returns: Formatter function to determine date/time format used.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._date_formatter

    @date_formatter.setter
    @class_sensitive(CallbackFunction)
    def date_formatter(self, value):
        self._date_formatter = value

    @property
    def describe_null(self) -> Optional[bool]:
        """If ``True``, will describe points with the value ``null`` to assistive
        technology (e.g. screen readers).

        Defaults to ``True``.

        :returns: Flag indicating whether to describe points with the value ``null`` to
          assistive technology.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._describe_null

    @describe_null.setter
    def describe_null(self, value):
        if value is None:
            self._describe_null = None
        else:
            self._describe_null = bool(value)

    @property
    def description_format(self) -> Optional[str]:
        """A :term:`format string` to use instead of the default for 
        point descriptions.
        
        The context of the format string is the point instance, but as opposed to 
        :meth:`.value_description_format <highcharts_core.options.accessibility.point.AccessibilityPoint.value_description_format>`, this option replaces the entire description.
        
        Defaults to :obj:`None <python:None>`.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._description_format
    
    @description_format.setter
    def description_format(self, value):
        self._description_format = validators.string(value, allow_empty = True)

    @property
    def description_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript formatter function to use instead of the default for point
        descriptions.

        The JavaScript formatter function should receive one argument, ``point``,
        referring to the point to describe. Should return a string with the description of
        the point for a screen reader user. If ``false`` is returned, the default
        formatter will be used for that point.

        .. tip::

            Best practice is to use :meth:`Accessibility.value_description_format` instead
            if possible, as default functionality such as describing annotations will be
            preserved.

        :returns: The JavaScript formatter function to use for point descriptions.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._description_formatter

    @description_formatter.setter
    @class_sensitive(CallbackFunction)
    def description_formatter(self, value):
        self._description_formatter = value

    @property
    def value_decimals(self) -> Optional[int]:
        """Number of digits after the decimal point to use for the values in the point
        description. Uses :meth:`Tooltip.value_decimals` if not provided here.

        :returns: Number of digits to use after the decimal point.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`

        :raises ValueError: if a negative number is provided
        """
        return self._value_decimals

    @value_decimals.setter
    def value_decimals(self, value):
        self._value_decimals = validators.integer(value,
                                                  allow_empty = True,
                                                  coerce_value = True,
                                                  minimum = 0)

    @property
    def value_description_format(self) -> Optional[str]:
        """Format to use for describing the values of data points to assistive technology
        - including screen readers. The point context is available as ``{{point}}``.

        Defaults to ``{xDescription}{separator}{value}``

        Other available context variables include ``{{index}}``, ``{{value}}``, and
        ``{{xDescription}}``.

        .. note::

          Additionally, the series name, annotation info, and description added in
          :meth:`Point.accessibility.description` is added by default if relevant.
          To override this, use the :meth:`AccessibilityPoint.description_formatter`
          option.

        :returns: Format to use for describing the values of data points to assistive
          technology.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_description_format

    @value_description_format.setter
    def value_description_format(self, value):
        self._value_description_format = validators.string(value, allow_empty = True)

    @property
    def value_prefix(self) -> Optional[str]:
        """Prefix to add to the values in the point descriptions.

        .. note::

          Uses :meth:`Tooltip.value_prefix` if not defined.

        :returns: Prefix to add to the values in the point description.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_prefix

    @value_prefix.setter
    def value_prefix(self, value):
        self._value_prefix = validators.string(value, allow_empty = True)

    @property
    def value_suffix(self) -> Optional[str]:
        """Suffix to add to the values in the point descriptions.

        .. note::

          Uses :meth:`Tooltip.value_suffix` if not defined.

        :returns: Suffix to add to the values in the point description.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value_suffix

    @value_suffix.setter
    def value_suffix(self, value):
        self._value_suffix = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'date_format': as_dict.get('dateFormat', None),
            'date_formatter': as_dict.get('dateFormatter', None),
            'describe_null': as_dict.get('describeNull', None),
            'description_format': as_dict.get('descriptionFormat', None),
            'description_formatter': as_dict.get('descriptionFormatter', None),
            'value_decimals': as_dict.get('valueDecimals', None),
            'value_description_format': as_dict.get('valueDescriptionFormat', None),
            'value_prefix': as_dict.get('valuePrefix', None),
            'value_suffix': as_dict.get('valueSuffix', None)
        }
        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'dateFormat': self.date_format,
            'dateFormatter': self.date_formatter,
            'describeNull': self.describe_null,
            'descriptionFormat': self.description_format,
            'descriptionFormatter': self.description_formatter,
            'valueDecimals': self.value_decimals,
            'valueDescriptionFormat': self.value_description_format,
            'valuePrefix': self.value_prefix,
            'valueSuffix': self.value_suffix
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core.options.plot_options.area import AreaOptions


class BellCurveOptions(AreaOptions):
    """General options to apply to all Bell Curve series types.

    A bell curve is an areaspline series which represents the probability density
    function of the normal distribution. It calculates mean and standard deviation of
    the base series data and plots the curve according to the calculated parameters.

    .. figure:: ../../../_static/bellcurve-example.png
      :alt: Bell Curve Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._intervals = None

        self.intervals = kwargs.get('intervals', None)

        super().__init__(**kwargs)

    @property
    def intervals(self) -> Optional[int | float | Decimal]:
        """Defines the length of the bell curve, expressed as the number of standard
        deviations. Defaults to ``3``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._intervals

    @intervals.setter
    def intervals(self, value):
        if value is None:
            self._intervals = None
        else:
            self._intervals = validators.numeric(value,
                                                 allow_empty = False,
                                                 minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'intervals': as_dict.get('intervals', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'intervals': self.intervals
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from highcharts_core.options.plot_options.series import SeriesOptions


class ParetoOptions(SeriesOptions):
    """General options to apply to all Pareto series types.

    A pareto diagram is a type of chart that contains both bars and a line graph,
    where individual values are represented in descending order by bars, and the
    cumulative total is represented by the line.

    .. figure:: ../../../_static/pareto-example.png
      :alt: Pareto Example Chart
      :align: center

    """
    pass



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Box(HighchartsMeta):
    """Containing the position of a box that should be avoided by labels."""

    def __init__(self, **kwargs):
        self._bottom = None
        self._left = None
        self._right = None
        self._top = None

        self.bottom = kwargs.get('bottom', None)
        self.left = kwargs.get('left', None)
        self.right = kwargs.get('right', None)
        self.top = kwargs.get('top', None)

    @property
    def bottom(self) -> Optional[int | float | Decimal]:
        """Bottom Position

        :rtype: numeric
        """
        return self._bottom

    @bottom.setter
    def bottom(self, value):
        self._bottom = validators.numeric(value)

    @property
    def left(self) -> Optional[int | float | Decimal]:
        """Left Position

        :rtype: numeric
        """
        return self._left

    @left.setter
    def left(self, value):
        self._left = validators.numeric(value)

    @property
    def right(self) -> Optional[int | float | Decimal]:
        """Right Position

        :rtype: numeric
        """
        return self._right

    @right.setter
    def right(self, value):
        self._right = validators.numeric(value)

    @property
    def top(self) -> Optional[int | float | Decimal]:
        """Top Position

        :rtype: numeric
        """
        return self._top

    @top.setter
    def top(self, value):
        self._top = validators.numeric(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'bottom': as_dict.get('bottom', None),
            'left': as_dict.get('left', None),
            'right': as_dict.get('right', None),
            'top': as_dict.get('top', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'bottom': self.bottom,
            'left': self.left,
            'right': self.right,
            'top': self.top
        }

        return untrimmed


class SeriesLabel(HighchartsMeta):
    """Series labels are placed as close to the series as possible in a natural way,
    seeking to avoid other series.

    The goal of this feature is to make the chart more easily readable, like if a human
    designer placed the labels in the optimal position.
    """

    def __init__(self, **kwargs):
        self._boxes_to_avoid = None
        self._connector_allowed = None
        self._connector_neighbour_distance = None
        self._enabled = None
        self._format = None
        self._formatter = None
        self._max_font_size = None
        self._min_font_size = None
        self._on_area = None
        self._style = None

        self.boxes_to_avoid = kwargs.get('boxes_to_avoid', None)
        self.connector_allowed = kwargs.get('connector_allowed', None)
        self.connector_neighbour_distance = kwargs.get('connector_neighbour_distance',
                                                       None)
        self.enabled = kwargs.get('enabled', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.max_font_size = kwargs.get('max_font_size', None)
        self.min_font_size = kwargs.get('min_font_size', None)
        self.on_area = kwargs.get('on_area', None)
        self.style = kwargs.get('style', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'plotOptions.boxplot.dragDrop'

    @property
    def boxes_to_avoid(self) -> Optional[List[Box]]:
        """An array of boxes to avoid when laying out the labels.
        Each item has a :meth:`left <Box.left>`, :meth:`right <Box.right>`,
        :meth:`top <Box.top>`, and :meth:`bottom <Box.bottom>` property.

        :rtype: :class:`list <python:list>` of :class:`Box`
        """
        return self._boxes_to_avoid

    @boxes_to_avoid.setter
    @class_sensitive(Box, force_iterable = True)
    def boxes_to_avoid(self, value):
        self._boxes_to_avoid = value

    @property
    def connector_allowed(self) -> Optional[bool]:
        """Allow labels to be placed distant to the graph if necessary, and draw a
        connector line to the graph. Defaults to ``False``.

        .. warning::

          Setting this option to ``True`` may decrease the performance significantly,
          since the algorithm with systematically search for open spaces in the whole plot
          area. Visually, it may also result in a more cluttered chart, though more of the
          series will be labeled.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._connector_allowed

    @connector_allowed.setter
    def connector_allowed(self, value):
        if value is None:
            self._connector_allowed = None
        else:
            self._connector_allowed = bool(value)

    @property
    def connector_neighbour_distance(self) -> Optional[int | float | Decimal]:
        """If the label is closer than this to a neighbour graph, draw a connector.
        Defaults to ``24``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._connector_neighbour_distance

    @connector_neighbour_distance.setter
    def connector_neighbour_distance(self, value):
        self._connector_neighbour_distance = validators.numeric(value, allow_empty = True)

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enable the series labels for the series. Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def format(self) -> Optional[str]:
        """A format string for the label, with support for a subset of HTML.

        Variables are enclosed by curly brackets. Available variables are ``name``,
        ``options.xxx``, ``color``, and other members from the ``series`` object.
        Use this option also to set a static text for the label.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format each of the series' labels.

        The ``this`` keyword refers to the ``series`` object. By default the formatter is
        :obj:`None <python:None>` and the ``series.name`` is rendered.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def max_font_size(self) -> Optional[int | float | Decimal | constants.EnforcedNullType]:
        """For area-like series, allow the font size to vary so that small areas get a
        smaller font size. The default applies this effect to area-like series but not
        line-like series.

        Defaults to ``EnforcedNull``.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._max_font_size

    @max_font_size.setter
    def max_font_size(self, value):
        if isinstance(value, constants.EnforcedNullType):
            self._max_font_size = constants.EnforcedNull
        else:
            self._max_font_size = validators.numeric(value, allow_empty = True)

    @property
    def min_font_size(self) -> Optional[int | float | Decimal | constants.EnforcedNullType]:
        """For area-like series, allow the font size to vary so that small areas get a
        smaller font size. The default applies this effect to area-like series but not
        line-like series.

        Defaults to ``EnforcedNull``.

        :rtype: numeric or :class:`EnforcedNullType` or :obj:`None <python:None>`
        """
        return self._min_font_size

    @min_font_size.setter
    def min_font_size(self, value):
        if isinstance(value, constants.EnforcedNullType):
            self._min_font_size = constants.EnforcedNull
        else:
            self._min_font_size = validators.numeric(value, allow_empty = True)

    @property
    def on_area(self) -> Optional[bool | constants.EnforcedNullType]:
        """Draw the label on the area of an area series.

        By default it is drawn on the area. Set it to ``False`` to draw it next to the
        graph instead.

        :rtype: :class:`bool <python:bool>` or :class:`EnforcedNullType` or
          :obj:`None <python:None>`
        """
        return self._on_area

    @on_area.setter
    def on_area(self, value):
        if value is None:
            self._on_area = None
        elif value is False:
            self._on_area = False
        elif isinstance(value, constants.EnforcedNullType):
            self._on_area = constants.EnforcedNull
        else:
            self._on_area = bool(value)

    @property
    def style(self) -> Optional[dict | str]:
        """Styles for the series label. The color defaults to the series color, or a
        contrast color if ``on_area``.

        :rtype: :class:`dict <python:dict>` or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        try:
            self._style = validators.dict(value, allow_empty = True)
        except (ValueError, TypeError):
            self._style = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'boxes_to_avoid': as_dict.get('boxesToAvoid', None),
            'connector_allowed': as_dict.get('connectorAllowed', None),
            'connector_neighbour_distance': as_dict.get('connectorNeighbourDistance',
                                                        None),
            'enabled': as_dict.get('enabled', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'max_font_size': as_dict.get('maxFontSize', None),
            'min_font_size': as_dict.get('minFontSize', None),
            'on_area': as_dict.get('onArea', None),
            'style': as_dict.get('style', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'boxesToAvoid': self.boxes_to_avoid,
            'connectorAllowed': self.connector_allowed,
            'connectorNeighbourDistance': self.connector_neighbour_distance,
            'enabled': self.enabled,
            'format': self.format,
            'formatter': self.formatter,
            'maxFontSize': self.max_font_size,
            'minFontSize': self.min_font_size,
            'onArea': self.on_area,
            'style': self.style
        }

        return untrimmed



from highcharts_core.options.plot_options.series import SeriesOptions


class SplineOptions(SeriesOptions):
    """General options to apply to all Spline series types.

    A spline series is a special type of line series, where the segments between the
    data points are smoothed.

    .. figure:: ../../../_static/spline-example.png
      :alt: Spline Example Chart
      :align: center

    """
    pass



from typing import Optional, List
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors, utility_functions
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class RotationOptions(HighchartsMeta):
    """Configuration of the allowed rotations for words in the wordcloud."""

    def __init__(self, **kwargs):
        self._from = None
        self._orientations = None
        self._to = None

        self.from_ = kwargs.get('from_', None)
        self.orientations = kwargs.get('orientations', None)
        self.to = kwargs.get('to', None)

    @property
    def from_(self) -> Optional[int | float | Decimal]:
        """The smallest degree of rotation allowed for a word. Defaults to ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._from

    @from_.setter
    def from_(self, value):
        self._from = validators.numeric(value,
                                        allow_empty = True,
                                        minimum = 0)

    @property
    def orientations(self) -> Optional[int]:
        """The number of possible orientations for a word, within the range of
        :meth:`RotationOptions.from_` and :meth:`RotationOptions.to`. Defaults to ``2``.

        .. warning::

          Must be a number larger than 0.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._orientations

    @orientations.setter
    def orientations(self, value):
        self._orientations = validators.integer(value,
                                                allow_empty = True,
                                                minimum = self.from_,
                                                maximum = self.to)

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The largest degree of rotation allowed for a word. Defaults to ``90``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value,
                                      allow_empty = True,
                                      minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'from_': as_dict.get('from', None),
            'orientations': as_dict.get('orientations', None),
            'to': as_dict.get('to', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'from': self.from_,
            'orientations': self.orientations,
            'to': self.to
        }

        return untrimmed


class WordcloudOptions(GenericTypeOptions):
    """General options to apply to all Wordcloud series types.

    A word cloud is a visualization of a set of words, where the size and placement of
    a word is determined by how it is weighted.

    .. figure:: ../../../_static/wordcloud-example.png
      :alt: Wordcloud Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        # Copied From SeriesOptions
        self._animation_limit = None
        self._color_index = None
        self._color_key = None
        self._relative_x_value = None

        # Copied from BaseBarOptions
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._center_in_category = None
        self._color_by_point = None
        self._colors = None
        self._edge_width = None

        # Native to WordcloudOptions
        self._allow_extend_playing_field = None
        self._max_font_size = None
        self._min_font_size = None
        self._placement_strategy = None
        self._rotation = None
        self._spiral = None

        # Copied from SeriesOptions
        self.animation_limit = kwargs.get('animation_limit', None)
        self.color_index = kwargs.get('color_index', None)
        self.color_key = kwargs.get('color_key', None)
        self.relative_x_value = kwargs.get('relative_x_value', None)

        # Copied from BaseBarOptions
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.center_in_category = kwargs.get('center_in_category', None)
        self.color_by_point = kwargs.get('color_by_point', None)
        self.colors = kwargs.get('colors', None)
        self.edge_width = kwargs.get('edge_width', None)

        # Native to WordcloudOptions
        self.allow_extend_playing_field = kwargs.get('allow_extend_playing_field', None)
        self.max_font_size = kwargs.get('max_font_size', None)
        self.min_font_size = kwargs.get('min_font_size', None)
        self.placement_strategy = kwargs.get('placement_strategy', None)
        self.rotation = kwargs.get('rotation', None)
        self.spiral = kwargs.get('spiral', None)

        super().__init__(**kwargs)

    @property
    def allow_extend_playing_field(self) -> Optional[bool]:
        """If there is no space for a word on the playing field, then setting this option
        to ``True`` will allow the playing field to be extended to fit the word. If
        ``False``, then the word will be dropped from the visualization. Defaults to
        ``True``.

        .. error::

          This option is currently not decided to be published in the API, and is
          therefore marked as/considered private.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_extend_playing_field

    @allow_extend_playing_field.setter
    def allow_extend_playing_field(self, value):
        if value is None:
            self._allow_extend_playing_field = None
        else:
            self._allow_extend_playing_field = bool(value)

    @property
    def animation_limit(self) -> Optional[int | float | Decimal]:
        """For some series, there is a limit that shuts down initial animation by default
        when the total number of points in the chart is too high. Defaults to
        :obj:`None <python:None>`.

        For example, for a column chart and its derivatives, animation does not run if
        there is more than 250 points totally. To disable this cap, set
        ``animation_limit`` to ``float("inf")`` (which represents infinity).

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._animation_limit

    @animation_limit.setter
    def animation_limit(self, value):
        if value == float('inf'):
            self._animation_limit = float('inf')
        else:
            self._animation_limit = validators.numeric(value,
                                                       allow_empty = True,
                                                       minimum = 0)

    @property
    def border_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the border surrounding each column or bar. Defaults to
        ``'#ffffff'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        self._border_color = utility_functions.validate_color(value)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The corner radius of the border surrounding each column or bar. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The width of the border surrounding each column or bar. If
        :obj:`None <python:None>`, defaults to ``1`` when there is room for a border, but
        to ``0`` when the columns are so dense that a border would cover the next column.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def center_in_category(self) -> Optional[bool]:
        """If ``True``, the columns will center in the category, ignoring null or missing
        points. When ``False``, space will be reserved for null or missing points.
        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._center_in_category

    @center_in_category.setter
    def center_in_category(self, value):
        if value is None:
            self._center_in_category = None
        else:
            self._center_in_category = bool(value)

    @property
    def color_by_point(self) -> Optional[bool]:
        """When using automatic point colors pulled from the global colors or
        series-specific collections, this option determines whether the chart should
        receive one color per series (``False``) or one color per point (``True``).

        Defaults to ``True``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._color_by_point

    @color_by_point.setter
    def color_by_point(self, value):
        if value is None:
            self._color_by_point = None
        else:
            self._color_by_point = bool(value)

    @property
    def color_index(self) -> Optional[int]:
        """When operating in :term:`styled mode`, a specific color index to use for the
        series, so that its graphic representations are given the class name
        ``highcharts-color-{n}``.

        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._color_index

    @color_index.setter
    def color_index(self, value):
        self._color_index = validators.integer(value,
                                               allow_empty = True,
                                               minimum = 0)

    @property
    def color_key(self) -> Optional[str]:
        """Determines what data value should be used to calculate point color if
        :meth:`AreaOptions.color_axis` is used.

        .. note::

          Requires to set ``min`` and ``max`` if some custom point property is used or if
          approximation for data grouping is set to ``'sum'``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color_key

    @color_key.setter
    def color_key(self, value):
        self._color_key = validators.string(value, allow_empty = True)

    @property
    def colors(self) -> Optional[List[str | Gradient | Pattern]]:
        """A series-specific or series type-specific color set to apply instead of the
        global colors when :meth:`BarOptions.color_by_point` is ``True``.

        :rtype: :class:`list <python:list>` of :class:`str <python:str>`,
          :class:`Gradient`, or :class:`Pattern` OR :obj:`None <python:None>`
        """
        return self._colors

    @colors.setter
    def colors(self, value):
        if not value:
            self._colors = None
        else:
            value = validators.iterable(value)

            self._colors = [utility_functions.validate_color(x) for x in value]

    @property
    def edge_width(self) -> Optional[int | float | Decimal]:
        """The width of the colored edges applied to a 3D column. Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._edge_width

    @edge_width.setter
    def edge_width(self, value):
        self._edge_width = validators.numeric(value,
                                              allow_empty = True,
                                              minimum = 0)

    @property
    def max_font_size(self) -> Optional[int | float | Decimal]:
        """The word with the largest weight will have a font size equal to this value.
        Defaults to ``25``.

        .. note::

          The font size of a word is the ratio between its weight and the largest occuring
          weight, multiplied with the value of ``max_font_size``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._max_font_size

    @max_font_size.setter
    def max_font_size(self, value):
        self._max_font_size = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def min_font_size(self) -> Optional[int | float | Decimal]:
        """A threshold determining the minimum font size that can be applied to a word.
        Defaults to ``1``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._min_font_size

    @min_font_size.setter
    def min_font_size(self, value):
        self._min_font_size = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @property
    def placement_strategy(self) -> Optional[str]:
        """This option decides which algorithm is used for the placement and rotation of
        a word in the wordcloud. Defaults to ``'center'``.

        .. note::

          It is possible for users to add their own custom placement strategies as
          described in the
          `JavaScript documentation <https://www.highcharts.com/docs/chart-and-series-types/word-cloud-series#custom-placement-strategies>`__

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._placement_strategy

    @placement_strategy.setter
    def placement_strategy(self, value):
        self._placement_strategy = validators.string(value, allow_empty = True)

    @property
    def rotation(self) -> Optional[RotationOptions]:
        """Configuration of the allowed rotations for words in the wordcloud.

        :rtype: :class:`RotationOptions` or :obj:`None <python:None>`
        """
        return self._rotation

    @rotation.setter
    @class_sensitive(RotationOptions)
    def rotation(self, value):
        self._rotation = value

    @property
    def spiral(self) -> Optional[str]:
        """Spiralling algorithm used for placing a word after the initial position
        experienced a collision with either another word or the borders. Defaults to
        ``'rectangular'``.

        .. note::

          It is possible for users to add their own custom spiralling algorithms for use
          in wordclouds. For more information, please see the
          `JavaScript documentation <https://www.highcharts.com/docs/chart-and-series-types/word-cloud-series#custom-spiralling-algorithm>`__

        :rtype: :class:`str <python:str>`
        """
        return self._spiral

    @spiral.setter
    def spiral(self, value):
        self._spiral = validators.string(value, allow_empty = True)

    @property
    def relative_x_value(self) -> Optional[bool]:
        """When ``True``, X values in the data set are relative to the current
        :meth:`point_start <AreaOptions.point_start>`,
        :meth:`point_interval <AreaOptions.point_interval>`, and
        :meth:`point_interval_unit <AreaOptions.point_interval_unit>` settings. This
        allows compression of the data for datasets with irregular X values. Defaults to
        ``False``.

        The real X values are computed on the formula ``f(x) = ax + b``, where ``a`` is
        the :meth:`point_interval <AreaOptions.point_interval>` (optionally with a time
        unit given by :meth:`point_interval_unit <AreaOptions.point_interval_unit>`), and
        ``b`` is the :meth:`point_start <AreaOptions.point_start>`.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._relative_x_value

    @relative_x_value.setter
    def relative_x_value(self, value):
        if value is None:
            self._relative_x_value = None
        else:
            self._relative_x_value = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            # Copied from SeriesOptions
            'animation_limit': as_dict.get('animationLimit', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'relative_x_value': as_dict.get('relativeXValue', None),

            # Copied from BaseBarOptions
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'center_in_category': as_dict.get('centerInCategory', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'colors': as_dict.get('colors', None),
            'edge_width': as_dict.get('edgeWidth', None),

            # Native to WordcloudOptions
            'allow_extend_playing_field': as_dict.get('allowExtendPlayingField', None),
            'max_font_size': as_dict.get('maxFontSize', None),
            'min_font_size': as_dict.get('minFontSize', None),
            'placement_strategy': as_dict.get('placementStrategy', None),
            'rotation': as_dict.get('rotation', None),
            'spiral': as_dict.get('spiral', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'allowExtendPlayingField': self.allow_extend_playing_field,
            'animationLimit': self.animation_limit,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'centerInCategory': self.center_in_category,
            'colorByPoint': self.color_by_point,
            'colorIndex': self.color_index,
            'colorKey': self.color_key,
            'colors': self.colors,
            'edgeWidth': self.edge_width,
            'maxFontSize': self.max_font_size,
            'minFontSize': self.min_font_size,
            'placementStrategy': self.placement_strategy,
            'relativeXValue': self.relative_x_value,
            'rotation': self.rotation,
            'spiral': self.spiral
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal
from collections import UserDict

from validator_collection import validators, checkers

from highcharts_core import utility_functions, constants
from highcharts_core.options.series.data.base import DataBase
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class ArcDiagramData(DataBase):
    """The definition of a data point for use in an :class:`ArcDiagramSeries`."""

    def __init__(self, **kwargs):
        self._from_ = None
        self._to = None
        self._weight = None

        self.from_ = kwargs.get('from_', None)
        self.to = kwargs.get('to', None)
        self.weight = kwargs.get('weight', None)

        super().__init__(**kwargs)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """The color for the individual link. Defaults to :obj:`None <python:None>`, which
        sets the link color the same as the node it extends from.

        .. hint::

          The ``series.fillOpacity`` option also applies to the points, so when setting a
          specific link color, consider setting the ``fillOpacity`` to 1.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._color

    @color.setter
    def color(self, value):
        if not value:
            self._color = None
        else:
            self._color = utility_functions.validate_color(value)

    @property
    def from_(self) -> Optional[str]:
        """The :meth:`ArcDiagramData.id` or :meth:`ArcDiagramData.name` of the node that
        the link runs from. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._from_

    @from_.setter
    def from_(self, value):
        self._from_ = validators.string(value, allow_empty = True)

    @property
    def to(self) -> Optional[str]:
        """The :meth:`ArcDiagramData.id` or :meth:`ArcDiagramData.name` of the node that
        the link runs to. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.string(value, allow_empty = True)

    @property
    def weight(self) -> Optional[int | float | Decimal]:
        """The weight of the link between the nodes. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._weight

    @weight.setter
    def weight(self, value):
        self._weight = validators.numeric(value, allow_empty = True)

    @classmethod
    def from_array(cls, value):
        if not value:
            return []
        elif checkers.is_string(value):
            try:
                value = validators.json(value)
            except (ValueError, TypeError):
                pass
        elif not checkers.is_iterable(value):
            value = [value]

        collection = []
        for item in value:
            if checkers.is_type(item, 'SinglePointData'):
                as_obj = item
            elif checkers.is_dict(item):
                as_obj = cls.from_dict(item)
            elif item is None or isinstance(item, constants.EnforcedNullType):
                as_obj = cls(y = None)
            elif checkers.is_numeric(item):
                as_obj = cls(y = item)
            elif checkers.is_iterable(item, forbid_literals = (str, bytes, dict, UserDict)):
                if len(item) == 3:
                    as_obj = cls(from_ = item[0],
                                 to = item[1],
                                 weight = item[2])
                else:
                    raise errors.HighchartsValueError(f'each data point supplied must either '
                                                    f'be an Arc Diagram Data Point or be '
                                                    f'coercable to one. Could not coerce: '
                                                    f'{item}')
            else:
                raise errors.HighchartsValueError(f'each data point supplied must either '
                                                  f'be an Arc Diagram Data Point or be '
                                                  f'coercable to one. Could not coerce: '
                                                  f'{item}')
            collection.append(as_obj)

        return collection

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        """Convenience method which returns the keyword arguments used to initialize the
        class from a Highcharts Javascript-compatible :class:`dict <python:dict>` object.

        :param as_dict: The HighCharts JS compatible :class:`dict <python:dict>`
          representation of the object.
        :type as_dict: :class:`dict <python:dict>`

        :returns: The keyword arguments that would be used to initialize an instance.
        :rtype: :class:`dict <python:dict>`

        """
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'color_index': as_dict.get('colorIndex', None),
            'custom': as_dict.get('custom', None),
            'description': as_dict.get('description', None),
            'events': as_dict.get('events', None),
            'id': as_dict.get('id', None),
            'label_rank': as_dict.get('labelrank', None),
            'name': as_dict.get('name', None),
            'selected': as_dict.get('selected', None),

            'from_': as_dict.get('from', None),
            'to': as_dict.get('to', None),
            'weight': as_dict.get('weight', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'from': self.from_,
            'to': self.to,
            'weight': self.weight,

            'accessibility': self.accessibility,
            'className': self.class_name,
            'color': self.color,
            'colorIndex': self.color_index,
            'custom': self.custom,
            'description': self.description,
            'events': self.events,
            'id': self.id,
            'labelrank': self.label_rank,
            'name': self.name,
            'selected': self.selected,
        }

        return untrimmed
